// module WaveClass10.cpp
// Reads, decodes, writes, plays and records WAVE files under Windows and Linux/ALSA
// Contains also methods for DFT (Discrete Fourier transformation) and 
// inverse DFT (note: not FFT!!!)
// Portable between Windows and Linux (this is a standard C++ program)
// Written by/Copyright Dr.E.Huckert 08-2021
//
// Restrictions:
// - only uncompressed formats (PCM)
// - only mono and stereo (1 oder 2 channels)

// versions:
// 10.01.2020: tested against FourierEHHarm.c: OK (same output)
//             write methods added
// 04.02.2021  write methods corrected, tested: OK
// 26.02.2021  methods makeWaveFormSine(), processInvDFT() added
// 01.03.2021  method processInvDFT() now OK, also tested under Linux
// 29.03.2021  addSamples(), allocateSampleData() methods added
//             include file waveclass10.h created
// 16.04.2021  getEpochTimeMs() renamed: getCurrentTimeMsMs()
// 03.05.2021  addSamples() for 8-bit samples added
//             constructor added with argument for bitsPerSample
// 06.05.2021  enhanced for writing of 8-bit samples 
//             bytes_per_second corrected
//             blockAlign corrected for 8-bit samples
// 23.07.2021  method play() added for WIN32 only, test is OK
// 30.07.2021  laeuft unter Windows und Linux/Ubuntu
// 01.08.2021  denormalizeData() redesigned
//             makeWaveFormSquare() redesigned
//             writeDataChunk() corrected, tested for stereo and 8/16 bit data
// 07.08.2021  thread (pThread based) added for keyboard detection and
//             overflow control (WAVE::record())
// 18.08.2021  WAVE::record() is now OK under Windows (WIN32) (V1.6)
//             Double buffering works! Interrupt via keyboard is OK.
// 19.08.2021  Problems fixed with casting of thread IDs in printf()
//             Routine resetLinuxTerminal() added
// 30.08.2021  Fixes in WAVE::writeDataChunk()
//             WAVE::record() in ALSA version is now OK
// 19.09.2021  makeWaveFormSine(): confusion between noSamples and 
//             noFrames eliminated, same for makeWaveFormSquare()
//             blockAlign corrected in constructors
// 21.10.2021  ::verbose replaced by this->verbose
// 05.11.2021: processDFTSamples() takes now sample no. instead
//             of millsecs
// 15.11.2021  readDataChunk() simplified
// 20.11.2021  Fix in WAVE::convertSamplesToShort() for stereo
// 21.11.2021  Fix in WAVE::play() WAVEHDR buffer length (for WINDOWS)
// 25.11.2021  8-bit samples are now played correctly under Linux/Alsa
// 26.11.2021  ALSA destructor enhanced: closes now the device
//             writeDataChunk() reformulated
// 01.12.201   convertSamplesToShort(): division by 1.15 added to avoid
//             overflow/underflow in conversion

// Test status 
// 18.11.2021
// OK: read WAVE files with 16-bit (signed) samples (WIN32 and Linux)
//     play these files (WIN32 and Linux)
//     normalize() and denormalize on these files (see test1())
// 20.11.2021
// OK: read/decode/play 8-Bit files, mono and stereo, see sound30.zip
// problem: playing of files is 50% to short
//          Linux: still problems playing 8 Bit samples
// 21.11.2021
// OK: duration of play corrected, see sound31.zip
// 25.11.2021
// OK: play 8-bit data under Linux/Alsa, OK also 16bit mono and 16-bit stereo
// 01.12.2021
// OK: normalization and denormalization (see test2())
// OK: write WAVE file for 16 bit output samples (Windows + Linux)
// OK: write WAVE file for 8-Bit output samples (Windows + Linux)

// compile:
// Windows: 
//          g++ -DWIN32 -DSTANDALONE -m32 -static -o WaveClass10.exe WaveClass10.cpp -lwinmm -lpthread
// Linux (GNU C): 
//          g++ -m32 -DSTANDALONE -static -o WaveClass10 WaveClass10.cpp -lasound -lpthread
//   Notes: on some Linux computers -static won't work(libasound problem)
//          -m32 may also make problems (then don't specify it)

// General note for WAVE files:
// "The default byte ordering assumed for WAVE data files is little-endian. 
// Files written using the big-endian byte ordering scheme have the 
// identifier F::READ instead of RIFF."
// EH: even if the data are always little endian the machine on which this
//     program runs may be big endian! This must be taken into account!

#include "waveclass10.h"
#define PORTION_SIZE 512  // no. of samples to be read/written per portion

#include <pthread.h>
static int threadCreated = 0;
static pthread_t threadId; 
static pthread_mutex_t mutex1;
static unsigned eventMask = 0;  // 1=stop, 2,4,8 WIM_...events

int verbose = 0;

// ----------------------------------------------------------
// sleep a number of milliseconds
void sleepMs(int millis)
{
#ifdef WIN32
  ::Sleep(millis);
#else
  ::usleep((long)millis * 1000L);
#endif
}   // end milliMs()

#ifdef WIN32
#include <winuser.h>
// ----------------------------------------------------------
// WIN32 version
int getKeyStroke() 
{
  //#define KEYPRESSED -32768
  short keystate = 0;
  int ok = 0;
  //
  // check only the ESC key
  keystate = ::GetAsyncKeyState(VK_ESCAPE);
  if (keystate & 0x8000)
    ok = 1;
  //
  return ok;
}   // end getKeyStroke()

// ---------------------------------------------------
// Only for WIN32!!!
char * reportLastError(DWORD dwError)
{
  //DWORD dwError  = ::GetLastError();
  //
  const DWORD dwFormatControl = FORMAT_MESSAGE_ALLOCATE_BUFFER |
                                FORMAT_MESSAGE_IGNORE_INSERTS |
                                FORMAT_MESSAGE_FROM_SYSTEM;
  LPVOID pTextBuffer = NULL;
  DWORD dwCount = ::FormatMessage(dwFormatControl, 
                                  NULL, 
                                  dwError, 
                                  0, 
                                  (LPTSTR)&pTextBuffer, 
                                  0, 
                                  NULL);
  return (char *)pTextBuffer;
}   // end reportLastError()
#else
#include <termios.h>
#include <unistd.h>
// -----------------------------------------------------------
// As we detect keystrokes interpreted as interrupt and
// change the terminal characterist we mus repair that...
void resetLinuxTerminal()
{
  // as we have changed the keyboard characteristics
  // can be different on different systems....
  //::system("stty sane erase 0x08");
  ::system("stty sane");
}   // end resetLinuxTerminal()

// ------------------------------------------------------------
// Linux version
// Sets the keyboard channel (0=stdin) to RAW mode (unbuffered)
// Returns: < 0 upon error, else the character read (or 0 if nothing read)
int getKeyStroke() 
{
  char buf           = 0;
  int  ret           = 0;
  static struct termios old = {0};
  static unsigned long entryCount = 0L;
  //
  if (entryCount == 0UL)
  {
    if (::tcgetattr(0, &old) < 0)
      ::perror("tcsetattr()");
    old.c_lflag     &= ~ICANON;
    old.c_lflag     &= ~ECHO;
    old.c_cc[VMIN]  = 1;
    old.c_cc[VTIME] = 0;
    // Note: must be reset later!!!
    if (tcsetattr(0, TCSANOW, &old) < 0)
      ::perror("tcsetattr ICANON");
  }
  //
  ret = ::read(0, &buf, 1);
  entryCount++;
  if (ret > 0)
    ret = buf & 0xff;
  return ret;
}   // end getKeyStroke()
#endif

// -------------------------------------------------
// thread worker routine
// Detects keyboard actions - calls getKeyStroke(), 
// sends a (pseudo) signal if necessary
static void *threadWorker1(void *pArg)
{
   long *pLong = NULL;
   unsigned long count = 0L;
   int ret = 0;
   //
   pLong = (long *)pArg;
   //
   if (::verbose)
   {
     if (pLong != NULL)
       ::printf("threadWorker1() arg=%ld\n", *pLong);
     else
       ::printf("threadWorker1() arg is NULL\n");
   }
   while (1)
   {
     if ((::verbose > 1) && ((count % 10) == 0))
       ::printf("threadWorker1() count=%lu\n", count);
     //
     ret = getKeyStroke();
     if (ret > 0)
     {
       // keyboard was activated
       if (::verbose)
         ::printf("threadworker1(): key activated\n"); 
       // 
       int rc = ::pthread_mutex_lock(&mutex1);
       if (rc != 0)
       {
         // cannot lock the mutex
         continue;
       }
       // set the stop signal
       ::eventMask |= 1;
       if (::verbose)
         printf("threadWorker1() KEYBOARD interrupt signal set\n");
       ::pthread_mutex_unlock(&mutex1);  
      }   // end if (ret > 0)       
      //
      ::sleepMs(100);
      count++;
      ::pthread_testcancel();   // this is necessary!!!
   }   // end while (1);
   //
   if (::verbose)
     ::printf("threadWorker1() leaving\n");
   ::pthread_exit(NULL);
}   // end threadWorker1()

// ---------------------------------------------------------
// create and start a thread
// Side effect: places the thread id in param. threadId
// Returns 0 if OK, < 0 if error
static int createThread(pthread_t *threadId,    
                 void *(*pWorker)(void *),  // address of worker routine
                 void *pArg)                 // NULL allowed
{
  int ret = 0;
  if (::verbose)
    ::printf("createThread() start\n");
  ret = ::pthread_create(threadId, 
                       NULL, 
                       pWorker,
                       pArg);  // argument passed to the worker routine
  if (ret != 0)
  {
    ret = -1;
    goto zurueck;
  }
  //
  zurueck:
  if (::verbose)
  #ifdef WIN32
    ::printf("createThread() threadId=%p\n", (*threadId));
  #else
    ::printf("createThread() threadId=%p\n", 
             (void *)(*threadId));
  #endif
  return ret;
}   // end createThread()

// -----------------------------------------------------
// Wait form the termination of a thread
// We are not interested in the the result of the terminated thread
static void terminateThread(pthread_t threadId)
{
  int ret;
  if (::verbose)
    #ifdef WIN32
    ::printf("terminateThread() threadId=%p\n", threadId);
    #else
   ::printf("terminateThread() threadId=%p\n", (void *)threadId);
    #endif
  //ret = pthread_join(thread_ids[threadNo], NULL);
  ret = ::pthread_cancel(threadId);
  if (::verbose)
    #ifdef WIN32
    ::printf("terminateThread() terminated threadId=%p\n", threadId);
    #else
    ::printf("terminateThread() terminated threadId=%p\n", (void *)threadId);
    #endif
}   // end terminateThread()

#ifndef WIN32
// for Linux/Alsa only
// ------------------------------------------------------------------
// Note: should be called after(!) reading/decoding the WAVE file
// Returns: 0 if OK, < 0 if error
int ALSA::open(WAVE *pWave,
               int mode)    // 0 = play, 1 = record (no duplex allowed)
{
  int ret  = 0;
  int cRet = 0;
  static int err, dir;
  static unsigned int rate           = 0;
  static snd_pcm_format_t format     = SND_PCM_FORMAT_S16_LE;
  static snd_pcm_hw_params_t *params = NULL;
  static unsigned long periodSz;
  snd_pcm_stream_t streamMode; 
  
  //
  if (this->verbose)
    ::printf("ALSA::open() start device=%s mode=%d\n", 
             (this->deviceName).c_str(), mode);
  // open() sets this->handle
  this->handle = NULL;
  if (mode == 0)
    streamMode = SND_PCM_STREAM_PLAYBACK;
  else
    streamMode = SND_PCM_STREAM_CAPTURE;
  err = ::snd_pcm_open(&(this->handle), 
                       (this->deviceName).c_str(),
                       streamMode, 
                       0);
  if (err < 0)
  {
    ::printf("ALSA::open() ERROR: %s\n", ::snd_strerror(err));
    ret = -1;
    goto zurueck; 
  }
  if (this->handle == NULL)
  {
    printf("ALSA::open() ERROR invalid PCM handle\n");
    ret = -2;
    goto zurueck; 
  }
  //
  // Important: avoid complicated snd_pcm... calls here!
  // It won't work!
  if (mode == 0)
  {
    // TODO: conversion here is not complete!
    // translate and set the format
    if (pWave->getFormat() == 1)
      format = SND_PCM_FORMAT_S16_LE;
    if (pWave->getBitsSample() <= 8)
      format = SND_PCM_FORMAT_U8;
    err = ::snd_pcm_set_params(this->handle,
                               format,
                               SND_PCM_ACCESS_RW_INTERLEAVED,
                               (unsigned int)(pWave->getNoChannels()),
                               (unsigned int)(pWave->getSamplesSec()),
                               // (unsigned int)(pWave->getSamplesSec() * pWave->getNoChannels()),
                               1, // 0,
                               500000);
    if (err < 0)
    {
      ::printf("ALSA::open() snd_pcm_set_params() error: %s\n", 
               ::snd_strerror(err));
      ret = -7;
      goto zurueck; 
    }
  }
  //
  zurueck:
  if (this->verbose)
    ::printf("ALSA::open() ret=%d\n", ret);
  return ret;
}   // end ALSA::open()

// ----------------------------------------------------------------------
// Destructor
// Side effect: closes the ALSA device
ALSA::~ALSA()
{
  if (this->verbose)
    ::printf("ALSA::~ALSA() destructor\n");
  if (this->handle != NULL)
    this->close();
  this->handle = NULL;
}

// ----------------------------------------------------------------------
// Returns: 0 if OK, < 0 if error
int ALSA::close()
{
   int ret = 0;
   //
   if (this->verbose)
     ::printf("ALSA::close()\n");
   if (this->getHandle() == NULL)
      ret = -1;
   else
   {
     ::snd_pcm_close(this->getHandle());
     this->handle = NULL;
   }
   if (this->verbose)
     ::printf("ALSA::close() ret=%d\n", ret);
   return ret;
 }
#endif

// ----------------------------------------------------------------------
// Returns the current time in ms since system start
// Note: ftime() replaced here by clock_gettime() since ftime() is deprecated
unsigned long getCurrentTimeMs()
{
  unsigned long timeMs = 0L;
  #ifdef WIN32
  timeMs = ::GetCurrentTime();
  #else
  #include <time.h>
  //#include <sys/timeb.h>
  struct timespec sNow;
  ::clock_gettime(CLOCK_REALTIME, &sNow);
  // note: seconds in sNow.tv_sec not considered
  timeMs = sNow.tv_nsec / 1000000;
  #endif
  return timeMs;
}   // end getCurrentTimeMs()

// ----------------------------------------------------------
// Constructor 1
WAVE::WAVE()
{
  //if (this->verbose)
  //{  ::printf("WAVE::WAVE() 1\n"); fflush(stdout); }
  this->basicInit();
}   // end WAVE::WAVE

// ----------------------------------------------------------
// Constructor 2
WAVE::WAVE(unsigned  sampsSec,
           unsigned  ds,  // duration in seconds
           unsigned  nc)  // no. of channels
{
  //if (this->verbose)
  //  ::printf("WAVE::WAVE() 2\n");
  this->basicInit();
  this->samplesSec = sampsSec;
  this->durSecs    = ds;
  this->noChannels = nc;
  this->blockAlign = sizeof(short) * this->noChannels;
}   // end WAVE::WAVE

// ----------------------------------------------------------
// Constructor 3
WAVE::WAVE(unsigned  sampsSec,
           unsigned  ds,  // duration in seconds
           unsigned  nc,  // no. of channels
           unsigned  bs)  // bits per sample
{
  //if (this->verbose)
  //  ::printf("WAVE::WAVE() 3\n");
  this->basicInit();
  this->samplesSec = sampsSec;
  this->durSecs    = ds;
  this->noChannels = nc;
  this->bitsSample = bs;
  if (bs <= 8)
    this->blockAlign = 1; 
  else
    this->blockAlign = sizeof(short) * this->noChannels;
}   // end WAVE::WAVE

// ----------------------------------------------------------
void WAVE::basicInit()
{
  this->ifd            = -1;
  this->verbose        = 0;
  this->noSamples      = 0L;
  this->samplesSec     = 44100; // CD quality
  this->noChannels     = 1;     // standard=Mono
  this->bitsSample     = 16;    // unsigned short 
  this->blockAlign     = sizeof(short) * this->noChannels;     // 
  this->bytesSec       = this->samplesSec * this->noChannels * ((this->bitsSample + 7) / 8);
  this->durSecs        = 0;     // duration in seconds
  this->frequency      = 0;
  #ifdef WIN32
  this->format         = WAVE_FORMAT_PCM;
  #else
  this->format         = SND_PCM_FORMAT_S16_LE; 
  #endif  
  this->pInpData       = NULL;
  this->noAllocatedSamples = 0L;  // the number of  doubles in pInpData
  this->pSineData      = NULL;    // precalculated sine values
  this->pSineTable     = NULL;
  this->pResults       = NULL;
  this->sineTabMax     = 360;
  this->resCount       = 0;
  this->maxResults     = 1;
  this-> freqStart     = 100; 
  this-> freqEnd       = 1000;
  this-> freqStep      = 20;
  this->normalized     = 0;
  this->normMaxVal     = 0.9;
  this->normMinVal     = -0.9;
  this->filename[0]    = 0;
}   // end WAVE::basicInit()

// ----------------------------------------------------------
// Destructor
WAVE::~WAVE()
{
  if (this->verbose)
    ::printf("WAVE::~WAVE()\n");
  if (this->ifd >= 0)
    close(this->ifd);
  this->freeSampleData();
  this->freeSineData();
  if (this->pSineTable != NULL)
    delete this->pSineTable;
  /*
  // be careful: if one than one instance exists!
  if (::threadCreated)
  {
    ::terminateThread(threadId);
    ::threadCreated = 0;
  }
  */
}   // end WAVE::~WAVE

// ----------------------------------------------------------
void WAVE::freeSineData()
{
  if (this->pSineData != NULL)
    delete this->pSineData;
  this->pSineData = NULL;
}   // end WAVE::freeSineData()

// ----------------------------------------------------------
// Just for tests
void WAVE::protocolDataPortion(unsigned long noOutps,
                           const char *pMsg)
{
  ::printf("%s\n", pMsg);
  for (unsigned long n=0; n < noOutps; n++)
  {
    ::printf("  n=%lu val=%lf\n", n, this->pInpData[n]);
  }
}   

// ----------------------------------------------------------
// Returns: < 0 upon error
// TODO: check overflow in pInpData
int WAVE::addSamples(short *pSamples,
                     unsigned long noSamp) // samples, not frames!
{
  int ret = 0;
  unsigned long idx = 0;
  //
  if (this->verbose)
  {
    ::printf("WAVE::addSamples(): noSamp=%lu\n", noSamp);
    ::fflush(stdout);
  }
  if (this->pInpData == NULL)
  {
    ret = -1;
    goto zurueck;
  }
  while (noSamp-- > 0L)
  {
    this->pInpData[this->noSamples++] = (double)(pSamples[idx++]);
  }
  ret = 0;
  zurueck:
  return ret;
}  // end WAVE::addSamples()

// ----------------------------------------------------------
// Returns: < 0 upon error
// TODO: check overflow in pInpData
int WAVE::addSamples(unsigned char *pSamples, // 8 bits!
                     unsigned long noSamp)
{
  int ret = 0;
  unsigned long idx = 0;
  //
  if (this->verbose)
  {
    ::printf("WAVE::addSamples(): noSamp=%lu\n", noSamp);
    ::fflush(stdout);
  }
  if (this->pInpData == NULL)
  {
    ret = -1;
    goto zurueck;
  }
  while (noSamp-- > 0L)
  {
    this->pInpData[this->noSamples++] = (double)(pSamples[idx++]);
  }
  ret = 0;
  zurueck:
  return ret;
}  // end WAVE::addSamples()

// ----------------------------------------------------------
// Returns: < 0 upon error
// TODO: check overflow in pInpData
int WAVE::addSamples(double  *pSamples,
                     unsigned long noSamp)
{
  unsigned long idx = 0;
  int      ret = 0;
  //
  if (this->pInpData == NULL)
  {
    ret = -1;
    goto zurueck;
  }
  while (noSamp-- > 0L)
  {
    this->pInpData[this->noSamples++] = pSamples[idx++];
  }
  ret = 0;
  zurueck:
  return ret;
}  // end WAVE::addSamples()

// ----------------------------------------------------------
// Modify the amplitudes in place
unsigned long WAVE::changeAmplitude(int percent)
{
  unsigned long lRet = 0;
  double newVal = 0.0;
  double dPercent = (double)percent;
  double percFact = (double)0.01 * dPercent;
  unsigned long noData;
  //
  if (this->pInpData == NULL)
  {
    lRet = -1L;
    goto zurueck;
  }
  if (this->verbose)
    ::printf("WAVE::changeAmplitude() percent=%d percFact=%lf\n", 
              percent,percFact );
  noData = this->noSamples * this->noChannels;
  for (unsigned long n=0; n < noData; n++)
  {
    newVal = this->pInpData[n] * percFact;
    this->pInpData[n] = newVal;
  }
  lRet = noData;
  zurueck:
  if (this->verbose)
    ::printf("WAVE::changeAmplitude() ret=%ld\n", lRet); 
  return lRet;
}   // end WAVE::changeAmplitude()

// ----------------------------------------------------------
void WAVE::freeSampleData()
{
  if (this->pInpData != NULL)
    delete this->pInpData;
  this->pInpData = NULL;
}   // end WAVE::freeSampleData()

// ----------------------------------------------------------
// Note: param noData must reflect the number of channels!
void WAVE::allocateSampleData(unsigned long noData)
{
  if (this->pInpData != NULL)
    delete this->pInpData;
  this->pInpData = (double *)new double[noData];
  this->noAllocatedSamples = noData;  
}   // end WAVE::allocateSampleData()

// ----------------------------------------------------------
// Simplified WAVE file handling
// read 12 bytes  main chunk header
// Returns: the number of bytes read
//          < 0 if error
long WAVE::readMainChunk()
{
  int nb;
  char buf[20];
  char *caddr;
  MAIN_CHUNK header;
  int ret = 0;
  //
  nb = ::read(this->ifd, &header, sizeof(MAIN_CHUNK));
  if (nb != sizeof(MAIN_CHUNK))
  {
    printf("WAVE::readMainChunk() ERROR in header of main chunk\n");
    return -1;
  }
  ret += nb;
  ::memcpy(buf, (char *)(&(header.main_chunk)), 4);
  buf[4] = 0;
  if (this->verbose)
  {
    printf("WAVE::readMainChunk()Kennung:     %s\n", buf);
    printf("WAVE::readMainChunk()Laenge:      %ld\n", (long)(header.mlength));
  }
  if (strncmp(buf, "RIFF", 4) != 0)
  {
     fprintf(stderr, "WAVE::readMainChunk() ERROR wrong file tag (should be RIFF)\n");
     return -2; 
  }
  if (this->verbose)
    printf("WAVE:readMainChunk() ret=%d\n",  ret);
  return ret;
}   // end WAVE:readMainChunk()

// -------------------------------------------------------------
// Simplified WAVE file handling
// read 24 bytes  sub chunk header (format header)
// starts at byte #12
// Returns: the number of bytes read
//          < 0 if error
long WAVE::readSubChunk()
{
  int  nb;
  char buf[20];
  char kennung[5];
  int  noRead = 0;
  //
  if (this->verbose)
    ::printf("readSubChunk() start\n");
  noRead = ::read(this->ifd, 
                  (char *)(&(this->formatHeader)), 
                  sizeof(SUB_CHUNK));
  this->format     = this->formatHeader.format;
  this->noChannels = this->formatHeader.channels;
  this->blockAlign = this->formatHeader.block_align;
  this->bitsSample = this->formatHeader.bits_per_sample;
  this->bytesSec   = this->formatHeader.bytes_per_second;
  //
  if (this->verbose)
  {
    ::printf("Sub-Kennung: %s\n",    kennung); // "fmt "
    ::printf("Sub-Laenge:  %u\n",   (unsigned)(this->formatHeader.slength));   // 4 Bytes
    ::printf("Format:      %04x\n", (short)(this->formatHeader.format));       // 2 Bytes
    ::printf("Kanaele:     %02x\n", (short)(this->formatHeader.channels));     // 2 Bytes
    ::printf("Sample-Freq: %ld\n",  (long)(this->formatHeader.sample_fq));     // 4 bytes
    ::printf("Bytes Rate:  %u\n",   (unsigned)this->bytesSec);                 // 4 bytes
    ::printf("Block Align: %02u\n", (short)this->formatHeader.block_align);    // 2 bytes
    ::printf("Bits/Sample: %u\n",   (unsigned)(this->formatHeader.bits_per_sample)); // 2 bytes
    ::printf("WAVE::readSubChunk() ret=%d\n",   noRead);
  }
  if (this->verbose)
    ::printf("readSubChunk() end\n");
  return noRead;;
}   // end WAVE:readSubChunk()

// ---------------------------------------------------------------
// Simplified WAVE file handling
// Read 8 bytes  data chunk header + all data bytes
// The bytes are stored as doubles in the allocated array pInpData
// Returns: the number of bytes read (also 0)
//          < 0 if error
// Quote from:https:
// stackoverflow.com/questions/10731226/
// how-to-determine-if-8bit-wav-file-is-signed-or-unsigned-using-java-and-without
// "In the wav File, 8-bit samples are stored as unsigned bytes, ranging from 0 to 255. 
// The 16-bit samples are stored as signed integers in 2's-complement."
long WAVE::readDataChunk(DATA_CHUNK *pHeader)
{
  int        nb;
  long       ll;
  unsigned   dataIdx = 0;
  long       noRead = 0L;
  static short      readBuf16[PORTION_SIZE];   // "static" must be for g++
  static unsigned char readBuf8[PORTION_SIZE]; // "static" must be for g++
  static char table16[256];
  static int table16Initialized = 0;
  double     dValue = 0.0;
  int        iValue = 0;
  short      sValue = 0;
  char       aux[16];
  long       filePos = 0L;
  static unsigned frameCount = 0;
  //
  // read 8 bytes = 2 DWORDs
  filePos = ::lseek(this->ifd, 0L, SEEK_CUR);
  if (this->verbose)
    ::printf("\nWAVE:readDataChunk() start pos=%lu\n", filePos);
  nb = ::read(this->ifd, pHeader, sizeof(DATA_CHUNK)); 
  if (nb <= 0)
  {
    if (this->verbose)
      ::printf("WAVE:readDataChunk() nb=%d\n", nb);
    if (nb == 0)
      noRead = 0L;
    else
      noRead = -1L;
    goto zurueck;
  }
  noRead += nb;
  ::memcpy(aux, (char *)(&(pHeader->data_type)), 4);
  aux[4] = 0;
  frameCount++;
  if (this->verbose)
  {
    ::printf("WAVE:readDataChunk() frame count=%u\n", frameCount);
    ::printf("WAVE:readDataChunk() id: %s\n", aux);
    ::printf("WAVE:readDataChunk() length:  %ld\n", (long)(pHeader->dlength));
  }
  if (::strncmp(aux, "data", 4) != 0)
  {
    ::printf("WAVE:readDataChunk() ERROR: not a data chunk\n");
    noRead = -2L;
    goto zurueck;    
  }
  //
  // allocate an array of double for the input data (sound samples)
  if (this->pInpData == NULL)
  {
    // Note: as the length is in the formula the number of channels is also in the formula!
    this->noSamples = (pHeader->dlength) / 
                      (((this->formatHeader.bits_per_sample + 7) / 8));
    this->allocateSampleData(this->noSamples);
    if (this->verbose)
      ::printf("WAVE:readDataChunk() noSamples=%lu bitsPerSample=%u\n", 
               this->noSamples, this->getBitsSample());
  }
  if (this->pInpData == NULL)
  {
    ::fprintf(stderr, 
       "WAVE:readDataChunk(): ERROR cannot allocate sample memory\n");
    noRead = -3L;
    goto zurueck;
  }
  //    
  // Read here the data (input sound samples)
  // Store the data in array pInpData
  if (this->verbose)
  {
    ::printf("WAVE:readDataChunk() pheader->dlength=%ld\n", 
           (long)(pHeader->dlength));
  }
  filePos = ::lseek(this->ifd, 0L, SEEK_CUR);
  if (this->verbose)
    ::printf("WAVE:readDataChunk() data pos=%lu\n", filePos);
  //
  dataIdx = 0;
  if (this->getBitsSample() > 8)
    ll = this->noSamples * sizeof(short);
  else
    ll = this->noSamples;
  if (this->verbose)
    ::printf("WAVE:readDataChunk() ll=%lu\n", ll);
  //
  while (ll > 0L)  // ll = length in bytes
  {
    if (this->getBitsSample() >= 16)
    {
      nb = sizeof(readBuf16);
      nb = ::read(this->ifd, (char *)readBuf16, nb);
    }
    else
    {
      nb = sizeof(readBuf8);
      nb = ::read(this->ifd, (char *)readBuf8, nb);
    }
    if (nb == 0) 
      // probably EOF
      break; 
    if (nb < 0) 
    {
      ::printf("WAVE:readDataChunk() read ERROR\n");
      #ifdef WIN32
      DWORD errCode = ::GetLastError();
      reportLastError(errCode);
      #endif
      noRead = -1000;
      goto zurueck;
    }
    ll     -= nb;
    noRead += nb;
    if (this->getBitsSample() > 8)
    {
      // 9..16 Bit samples are signed shorts in input
      nb = nb / sizeof(short);
      for (int i = 0;    // i is index in shorts!
           i < nb; 
           i++)
      {
        // here only for 16 Bit signed data
        // convert the sample to a double value
        dValue = (double)(readBuf16[i]);
        // put the sample value in array pInpData
        this->pInpData[dataIdx] = dValue;
        dataIdx++;
      }   // end for i...
    }
    else
    {
      unsigned short idx = 0;
      // 8 Bit samples are unsigned bytes in input
      for (int i = 0;    // i is index in bytes!
           i < nb; 
           i++)
      {
        if (! table16Initialized)
        {
          // prepare the lookup table
          sValue = 127;
          for (int n= 255; n >= 0; n--)
          {
            table16[n] = sValue;
            sValue--;
          }
          table16Initialized = 1;
        }
        // here only for 8 Bit unsigned data
        // convert the sample to a double value
        idx = readBuf8[i];
        dValue = (double)(table16[idx]) * 255.0;
        // put the sample value in array pInpData
        this->pInpData[dataIdx] = dValue;
        dataIdx++;
      }   // end for i...
    }
  }   // end while (ll > 0L)
  //
  this->setNormalized(0);
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE:readDataChunk() ret=%ld\n", noRead);
  return noRead;
}   // end WAVE:readDataChunk()

// --------------------------------------------------------------
// dump data (samples) to stdout
// data are taken from the input file, not from the internal memory!
// The current file position is kept and restored
// The current WAVE file must be open for read
int WAVE::dumpPureData(long filePos,
                       unsigned long noSamps)
{
  int        ret = 0;
  int        nb = 0;
  unsigned   dataIdx = 0;
  long       oldFilePos = 0L;
  short      readBuf[PORTION_SIZE];
  short      sVal;
  int        iVal;
  unsigned short usVal;
  double     dVal;
  //
  oldFilePos = ::lseek(this->ifd, 0L, SEEK_CUR);
  filePos = ::lseek(this->ifd, filePos, SEEK_SET);
  while (noSamps > 0L)
  {
    nb = sizeof(readBuf);
    if (noSamps < nb) 
      nb = noSamps;
    nb = ::read(this->ifd, (char *)readBuf, nb);
    if (nb <= 0) 
    {
      // probably EOF
      ret = -1;
      goto zurueck;
    }
    noSamps -= nb / sizeof(short);
    for (int i = 0; 
         i < (nb / sizeof(short)); 
         i++)
    {
      dVal = (double)readBuf[i];
      iVal = (int)readBuf[i];
      sVal =  readBuf[i];
      usVal = (unsigned short)(readBuf[i]);
      ::printf("%05u %04x %05d %05u %05d %6.2lf\n", 
                dataIdx, (readBuf[i] & 0xffff),
                sVal, usVal, iVal, dVal);
      dataIdx++;
    }   // end for i...
  }   // end while (noSamps > 0L)
  //
  zurueck:
  // restore the old file position
  ::lseek(this->ifd, oldFilePos, SEEK_SET);
  ::fflush(stdout);
  return ret;
}   // end WAVE::dumpPureData()

// --------------------------------------------------------------
// Simplified WAVE file handling
// Read a sequence of DATA_CHUNK frames
// The real data are in chunks of type 'data'
long WAVE::readData(unsigned long *readPos)  // will be changed!
{
  long lnb         = 0L;
  long lRet        = 0L;
  DATA_CHUNK dataHeader;
  //
  ::lseek(this->ifd, 36L, SEEK_SET);  // EH added 26.8.2020
  while (1)
  {
    lnb = this->readDataChunk(&dataHeader);
    if (lnb <= 0L)
    {
      break;
    }
    *readPos += lnb;
    lRet     += lnb;
  }  // end while 1
  //
  if (this->verbose)
    ::printf("WAVE:readData() ret=%ld\n", lRet);
  return lRet;
}   // end WAVE:readData()

// ---------------------------------------------------------------
// Returns: 0L if OK
//          < 0L upon error
long WAVE::read(const char *fileName)
{
  long   lRet    = 0L;
  unsigned long readPos = 0L;
  long   nb      = 0L;
  //
  if (this->verbose)
    ::printf("WAVE::read() start filename=%s\n", fileName);
  ::strcpy(this->filename, fileName);
 #ifdef WIN32
  this->ifd = ::open(fileName, O_RDONLY + O_BINARY);
  #else
  this->ifd = ::open(fileName, O_RDONLY);  // Linux is always binary!
  #endif
  if (ifd < 0)
  {
    lRet = -1L;
    ::printf("WAVE::read() ERROR: Inp.wave file not found: %s\n", fileName);
    goto zurueck;
  }
  //
  // Read the first header: "RIFF", Laenge, "WAVE"
  nb = this->readMainChunk();
  if (nb < 0)
  {
    ::printf("WAVE::read() ERROR: main chunk error\n");
    lRet = -2L;
    goto zurueck;
  }
  readPos += nb;
  if (this->verbose)
  {
    ::printf("WAVE::read() after main chunk read pos=%ld 0x%lx\n", 
           (long)readPos, (long)readPos);
    ::printf("\n");
  }
  //
  // Read the second header (sub chunk, format chunk)
  // Contains the number of channels and the sample freq.
  nb = this->readSubChunk();
  if (nb < 0)
  {
    ::printf("WAVE::read() ERROR: sub chunk\n");
    lRet = -3L;
    goto zurueck;
  }
  readPos += nb;
  if (this->verbose)
  {
     ::printf("WAVE::read() %ld bytes read in readSubChunk()\n", nb);
     ::printf("WAVE::read() after readSubChunk() read pos=%ld 0x%lx\n", 
            (long)readPos, (long)readPos);
     ::printf("WAVE::read() noChannels=%u\n", this->noChannels);
  }
  // 
  // Read all data (all data chunks)
  // Note: this converts short data to double data!
  lRet =  this->readData(&readPos);
  if (lRet < 0)
    lRet = -4L;
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE::read() readPos=%ld ret=%ld\n", (long)readPos, lRet);
  return lRet;
}   // end WAVE::read()

// --------------------------------------------------------------
// Normalize the input data (Y values) to range -1 .. +1
// This is necessary as the Y sine/cosine values will also be in this range
unsigned long WAVE::normalizeData()
{
  unsigned long lRet = 0L;
  double maxVal = 0.1;
  double minVal = -0.1;
  unsigned count = 0;
  //
  if (this->verbose)
    printf("WAVE::normalizeData() no samples=%lu\n", this->noSamples);
  if (this->normalized)
  {
    // data are already normalized
    if (this->verbose)
      printf("WAVE::normalizeData() data are normalized - nothing to do\n");
    goto zurueck;
  }
  //
  // step 1: find the max. and min. values
  for (unsigned i = 0; i < this->noSamples; i++)
  {
    if (this->pInpData[i] > maxVal)  maxVal = pInpData[i];
    if (this->pInpData[i] <= minVal) minVal = pInpData[i];
  }
  if (maxVal == 0.0) maxVal = 0.1;  // avoid div. by 0
  if (minVal == 0.0) minVal = 0.1;  // avoid div. by 0
  if (this->verbose)
    printf("WAVE::normalizeData() minVal=%lf maxVal=%lf\n",
            minVal, maxVal);
  if (minVal < 0.0)  
    minVal = ::fabs(minVal);  // keep the sign in the results
  //
  // step 2: normalize for range -1 to +1
  for (unsigned i = 0; i < this->noSamples; i++)
  {
    if (this->pInpData[i] >= 0.0)
      this->pInpData[i] = this->pInpData[i] / maxVal;
    else
      this->pInpData[i] = this->pInpData[i] / minVal;
  }
  this->normalized = 1;
  // keep the max/min values for later denormalization
  this->normMaxVal = maxVal;
  this->normMinVal = minVal;
  lRet = this->noSamples;
  //
  zurueck:
  if (this->verbose)
    printf("WAVE::normalizeData() ret=%lu\n", lRet);
  return lRet;
}   // end WAVE:normalizeData()

// -----------------------------------------------------------------
void WAVE::setDenormalizeRange(double minVal, double maxVal)
{ 
  this->normMinVal = minVal; 
  this->normMaxVal = maxVal;
  if (this->verbose)
    printf("WAVE::setDenormalizeRange() min=%lf max=%lf\n", 
            this->normMinVal, this->normMaxVal);
}   // end WAVE::setDenormalizeRange()

// --------------------------------------------------------------
// Denormalize the input data (Y values) from  range -1.0 .. +1.0 into
// range 0xffff to 0x7fff (signed shorts).
// The denormalized data replace (overwrite) the old normalized data.
// The internal data will still be kept in double format
// Returns: the number of samples denormalized
// Note: we always assume here storage as double values/16 bit signed shorts
//       as we deal here with internal sample data
unsigned long WAVE::deNormalizeData()
{
  unsigned long lRet = 0L;
  unsigned count = 0;
  double   maxVal = 0.0;
  double   minVal = 0.0;
  double   oldData = 0.0;
  double   factor = 127.0;
  //
  if (this->verbose)
    printf("WAVE::deNormalizeData() noSamples=%lu bitsSample=%u\n", 
           this->noSamples, this->bitsSample);
  /*
  if (this->bitsSample <= 8)
     factor = 127.0 - 2.0; // -2.0: avoid clipping 
  if (this->bitsSample > 8)
  */
  factor = (double)(0x7fff - 10);  // -10: avoid clipping
  if (this->verbose)
    printf("WAVE::deNormalizeData() factor=%lf\n", factor);
  //
  // map -1:+1 to -32767:+32767
  for (unsigned long i = 0L; i < this->noSamples; i++)
  { 
    if (this->pInpData[i] > maxVal) maxVal = this->pInpData[i];
    if (this->pInpData[i] < minVal) minVal = this->pInpData[i];
    oldData = this->pInpData[i];
    this->pInpData[i] = this->pInpData[i] * factor;
    if (this->bitsSample <= 8)
    {
      // map as unsigned bytes
      this->pInpData[i] = this->pInpData[i] + 127.0;
    }
  }   // end for ...
  if (this->verbose)
    ::printf("WAVE::deNormalizeData() minVal=%lf maxVal=%lf\n", 
             minVal, maxVal);
  this->normalized = 0;
  lRet = this->noSamples;
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE::deNormalizeData() ret=%lu\n", lRet);
  return lRet;
}   // end WAVE:deNormalizeData()

// --------------------------------------------------------------
// Resample the data (samples) so that the speed is changed
// This replaces the original data with the changed data!
// A large number of new data must be allocated
// Returns: the number of new samples
// Note: aliasing effects with the algorithm below

#define SEGMENT_SIZE  20  // 25 or smaller!

unsigned long WAVE::resampleData(float percent)
{
  unsigned long lRet = 0L;
  unsigned long noNewSamples = 0L;
  double * pNewSamples = NULL;
  double   d1, d2, stretch;
  unsigned long jNew, jOld, dj;
  double   segment[SEGMENT_SIZE];
  //
  if (this->verbose)
    ::printf("WAVE::resampleData() no samples=%lu\n", this->noSamples);
  //
  // Allocate a new array
  try
  {
    d1 = ((double)this->noSamples) / (double)100.0;
    d2 = d1 * (double)percent;
    noNewSamples = this->noSamples + (unsigned long)d2;
    stretch = ((double)SEGMENT_SIZE / 100.0) * fabs(percent);
    stretch = (double)SEGMENT_SIZE + stretch;
    if (this->verbose)
      ::printf("WAVE::resampleData() no new samples=%lu stretch=%lf\n", 
                   noNewSamples, stretch);
    pNewSamples  = new double[noNewSamples];
    if (pNewSamples == NULL)  // OK for Digital Mars!
    {
      //throw (std::bad_alloc e);
      ::printf("WAVE::resampleData() EXCEPTION bad_alloc");
    }
    if (this->verbose)
      ::printf("WAVE::resampleData() array pNewSamples allocated\n");
    jNew = jOld = 0L;
    while (jOld < this->noSamples)
    {
      // fill a segment of old data
      for (unsigned long i=0L; i < SEGMENT_SIZE; i++)
      {
        if ((jOld + i) >= this->noSamples)
          goto enough;
        segment[i] = this->pInpData[jOld + i];
      }
      //
      // fill now the segment into the new samples array
      dj = 0.0;
      while (1)
      {
        for (unsigned long i=0L; i < SEGMENT_SIZE; i++)
        {
          if (jNew >= noNewSamples)
            goto enough;
          pNewSamples[jNew++] = segment[i];
          dj = dj + 1.0;
          if (dj >= stretch)
            goto nextSegment;
        }
      }
      nextSegment:
      jOld = jOld + SEGMENT_SIZE;
    }   // end while ...
    enough:
    if (this->verbose)
      ::printf("WAVE::resampleData() all data resampled\n");
    // OK here
    lRet            = noNewSamples;
    if (pInpData != NULL)
      delete this->pInpData;
    this->pInpData  = pNewSamples;
    this->noSamples = noNewSamples;
  }
  catch (const std::bad_alloc &e)
  {
    lRet = -1;
    ::printf("WAVE::resampleData() EXCEPTION bad_alloc");
    //if (this->verbose)
    //  ::printf("WAVE::resampleData() EXCEPTION: %s\n", e.what());
  }
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE::resampleData() ret=%lu\n", lRet);
  return lRet;
}   // end WAVE:resampleData()

// ----------------------------------------------------------
// Fill the array pSineData with sine values
// Array pSineData has the same X values the the array of input samples.
// Array pSineData is allocated here if == NULL
// Array pSineTable contains all sine values for X=0 ... 360 (degrees)
// Note: parameter maxX is the same for mono and stereo 
int WAVE::makeSineData(unsigned maxX,  // same for stereo and mono!
                       unsigned noHertz)  
{
  int      ret  = 0;
  double   step = 1.0;
  #define PI   ((double)3.1415926)
  double   radianFactor = PI / 180.0;
  double   sineValue = 0.0;
  unsigned n       = 0;
  double   doubleN = 0.0;
  //
  if (noHertz == 0)
    noHertz = 1;    // avoid division by zero
  if (this->verbose)
    ::printf("WAVE::makeSineData() maxX=%u noHertz=%u\n", 
            maxX, noHertz);
  if (this->pSineData != NULL)
    // free memory from last calculation
    delete this->pSineData;
  this->pSineData = NULL;  // 
  //
  // allocate an array of double for the sine values
  // On exit array pSineData will contain sine values for X values corresponding 
  // to X values in the input samples
  if (this->pSineData == NULL)
  {
    this->pSineData = new double[maxX]; 
    if (this->verbose)
      ::printf("WAVE::makeSineData() memory for pSineData allocated\n"); 
  }
  if (this->pSineData == NULL)
  {
    ::fprintf(stderr, "WAVE::makeSineData() ERROR cannot allocate memory for sine data\n");
    ret = -1;
  }
  //
  step = (double)this->noSamples/(double)noHertz; // = no.samples for 360 degrees
  step = step / (double)360.0;                 // = no_samples for 1 degree
  if (this->verbose)
      ::printf("WAVE::makeSineData() step=%lf\n", step);
  if (step <= 0.0) 
    step = 0.01;
  //
  // check if the standard sine table (x args from 0 to 360) is already filled
  int tableFilled = 0;
  if (this->pSineTable == NULL)
  {
    tableFilled = 0;
    this->pSineTable = new double[this->sineTabMax];
    if (this->verbose)
      ::printf("WAVE::makeSineData() memory allocated for pSineTable\n");
  }
  else
    tableFilled = 1;
  if (this->verbose)
      ::printf("WAVE::makeSineData() pSineTable filled=%d\n", tableFilled);
  //
  n       = 0;    // n=index in the samples (X value)
  doubleN = 0.0;
  //
  while (n < maxX) 
  {
    // 17.11.2020 optimized: use a fixed sine value table
    for (unsigned sineArg = 0; sineArg < this->sineTabMax; sineArg++)
    {
      n = (unsigned)doubleN;
      if (n >= maxX) 
      {
        // Note: this happens quite frequently - no problem!
        if (this->verbose)
          ::printf("WAVE::makeSineData() pSineData overflow at n=%u!!!!\n", n);
        goto zurueck;
      }
      if (tableFilled == 0)
      {
        // standard sine table is not yet filled - preset the sine table
        sineValue = (double)(sin((double)sineArg * radianFactor));
        this->pSineTable[sineArg] = sineValue;
      }
      else
      {
        // get the value from the filled standard sine table
        sineValue = this->pSineTable[sineArg];
      }
      this->pSineData[n] = sineValue;
      doubleN += step;
    }   // end for ...
  }   // end while ...
  //
  ret = 0;
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE::makeSineData() ret=%d\n", ret);
  return ret;
}   // end makeSineData()

// ----------------------------------------------------------
// Produces a sine wave and fills the main sample buffer
// Produces stereo or mono interleaved data (depending on this->noChannels)
// The sample buffer is freed and then allocated here
// The samples are normalized in range -1. to +1.0
// Returns: the number of values generated
long WAVE::makeWaveFormSine(unsigned hertz,
                            unsigned long nFrames)  // frames, not samples!
{ 
  double         actVal, arg1, x, period;
  double         pi = 3.14159;
  long           lRet = 0;
  double         piRad = pi / (double)180.0;
  //
  if (hertz == 0) 
    hertz = 1;
  this->freeSampleData();  // free the old data (samples)
  period          = (double)this->samplesSec / (double)hertz;
  if (this->verbose)
    ::printf("WAVE::makeWaveFormSine() hertz=%u nFrames=%ld period=%lf\n", 
              hertz, nFrames, period);
  this->noSamples = nFrames * this->noChannels;
  // allocate a new data buffer (for the samples)
  // for stereo: interleaved format assumed
  this->allocateSampleData(this->noSamples);
  if (this->pInpData == NULL)
  {
    lRet = -1;
    goto zurueck;
  }
  //
  x = 0.0;
  for (unsigned long i = 0L; i < nFrames; i += this->noChannels)
  {
    if (x >= period) x = 0.0;
    // Argument fuer sin() muss in radians ausgedrueckt werden
    // Der resultierende Wert liegt im Bereich (-1,+1]
    arg1 = (double)x * (360.0 / (double)period);
    actVal = sin(arg1 * piRad);
    unsigned j = i;
    for (unsigned m=0; m < this->noChannels; m++)
      this->pInpData[j++] = actVal;
    x += 1.0;
  }   // end for ...
  //
  this->frequency   = hertz;
  this->durSecs     = nFrames / this->samplesSec;
  this->normalized  = 1;
  lRet              = nFrames;
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE::makeWaveFormSine() ret=%ld\n", lRet);
  return lRet;
}   // end WAVE::makeWaveFormSine()

// ---------------------------------------------------------------------
// Produces a square wave and fills the main sample buffer
// Produces stereo or mono interleaved data (depending on this->noChannels)
// The sample buffer is freed and then allocated here
// The samples are normalized in range -1. to +1.0
// Returns: the number of values generated
long WAVE::makeWaveFormSquare(unsigned hertz,
                              unsigned long nFrames)  // frames, not samples!
{
  double         actVal, x, period;
  long           lRet = 0;
  int            flipFlop  = 0;
  double         minVal = 0.0;
  double         maxVal = 0.0;
  //
  if (hertz == 0) 
    hertz = 1;
  this->freeSampleData();  // free the old data (samples)
  period = (double)this->samplesSec / (double)hertz;
  if (this->verbose)
    ::printf("WAVE::makeWaveFormSquare() hertz=%u nFrames=%ld period=%lf\n", 
              hertz, nFrames, period);
  this->noSamples = nFrames * this->noChannels;
  //
  // allocate a new data buffer (for the samples)
  // for stereo: interleaved format assumed
  this->allocateSampleData(this->noSamples);
  if (this->pInpData == NULL)
  {
    lRet = -1;
    goto zurueck;
  }
  //
  x = 0.0;
  for (unsigned long i = 0L; i < nFrames; i += this->noChannels)
  {
    if (x >= period) 
    {
      x = 0.0;
      flipFlop++;
    }
    if ((flipFlop % 2) == 0)
      actVal = 1.0 - 0.01; // -0.01 to avoid clipping
    else
      actVal = -1.0 + 0.01; // +0.01 to avoid clipping
    unsigned j = i;
    for (unsigned m=0; m < this->noChannels; m++)
      this->pInpData[j++] = actVal;
    if (actVal > maxVal) maxVal = actVal;
    if (actVal < minVal) minVal = actVal;
    x += 1.0;
  }   // end for ...
  //
  this->frequency   = hertz;
  this->durSecs     = nFrames / this->samplesSec;
  this->normalized  = 1;
  lRet              = nFrames;
  //
  zurueck:
  if (this->verbose)
    ::printf("WAVE::makeWaveFormSquare() min=%lf max=%lf ret=%ld\n", 
              minVal, maxVal, lRet);
  return lRet;
}   // end WAVE::makeWaveFormSquare()

// ----------------------------------------------------------
// Process the DFT (Discrete Fourier Transformation)
// Uses the samples in pInpData
// Calls its own sine generation makeSineData()
// Multiplies the inp.values with the corresponding sine values
// Note: phase problems may arise as the sine values start always with index 0
// Returns: the absolute(!) average value of all products.
double WAVE::processDFTSamples(unsigned long fromSamp,
                               unsigned long noSamp,
                               unsigned noHertz) 
{
  double dRet = 0.0;
  double multVal = 0.0;
  double sumVal = 0.0;
  double *pLocalResults = NULL;
  unsigned long nn = 0L;
  static unsigned outCount = 0;
  //
  if (this->verbose)
    ::printf("WAVE::processDFTSamples() fromSamp=%lu noSamps=%lu frequency=%u\n", 
            fromSamp, noSamp, noHertz);
  //
  pLocalResults = new double[noSamp];  // array starts from 0!!!
  if (this->verbose)
    ::printf("WAVE::processDFTSamples() memory for pLocalResults allocated\n");
  //
  this->makeSineData((unsigned)noSamp, noHertz);
  if (this->pSineData == NULL)
  {
    if (this->verbose)
      ::printf("WAVE::processDFT() ERROR no sine data\n");
    goto zurueck;
  }
  //
  nn = 0L;
  for (unsigned long n=fromSamp; n < (fromSamp + noSamp -1); n++)
  {
    multVal = this->pInpData[n] * this->pSineData[nn];
    if (this->verbose > 1)
      ::printf("WAVE::processDFTSamples() sample=%lu multVal=%lf\n", 
              n, multVal);
    pLocalResults[nn] = ::fabs(multVal);
    nn++;
  }   // end for n ...
  //
  // Calculate the sum of all multipl.results for that frequency
  sumVal = 0.0;
  for (unsigned long n=0L; n < noSamp; n++)
  {
    sumVal = sumVal + pLocalResults[n];
  }
  // Divide the result by the num. of samples to get an average value
  if (sumVal == 0.0) sumVal = 0.001;  // avoid div. by zero
  dRet = sumVal / (double)noSamp;
  //
  zurueck:
  //if (this->verbose)
  //    ::printf("WAVE::processDFTSamples() freeing memory pLocalResults\n"); 
  if (pLocalResults != NULL)
    delete pLocalResults;
  if (this->verbose)
  {
    ::printf("WAVE::processDFTSamples() dRet=%lf\n", dRet);
    ::fflush(stdout);
  }
  return dRet;
}   // end WAVE::processDFTSamples()

// ----------------------------------------------------------
// Process the DFT (Discrete Fourier Transformation)
// Considers only the data range in the given ms frame
// Uses the samples in pInpData
// Sine calculation is done here!
// Multiplies the inp.values with the corresponding sine values.
double WAVE::processDFT(unsigned long fromMs,
                        unsigned long noMs,
                        unsigned      noHertz) 
{
  double ret = 0.0;
  double multVal = 0.0;
  double sumVal = 0.0;
  double *pLocalResults = NULL;
  unsigned long fromXp = 0L;
  unsigned long toXp   = 0L;
  double   samplesMs = 0.0;
  static unsigned outCount = 0;
  //
  if (this->verbose)
    ::printf("WAVE::processDFT(3args) fromMs=%lu noMs=%lu frequency=%u\n", 
             fromMs, noMs, noHertz);
  //
  samplesMs = (double)(this->samplesSec / (double)1000.0);
  fromXp = fromMs * samplesMs;
  toXp   = (fromMs + noMs) * samplesMs;
  //
  pLocalResults = new double[toXp - fromXp];
  if (this->verbose)
    ::printf("WAVE::processDFT(3args) memory for pLocalResults allocated\n");
  //
  if (this->verbose)
    ::printf("WAVE::processDFT() fromXP=%lu toXp=%lu\n",
              fromXp, toXp);
  //
  //this->makeSineData((unsigned)(toXp - fromXp), noHertz);
  //
  for (unsigned long n=fromXp; n < toXp; n++)
  {
    multVal = this->pInpData[n] * 
      //this->pSineData[n - fromXp];
      this->pSineData[n];             // avoid phase problems!!!!
    if (this->verbose > 1)
      ::printf("WAVE::processDFT() sample=%lu multVal=%lf\n", 
              n, multVal);
    pLocalResults[n - fromXp] = multVal;
  }   // end for n ...
  //
  // Calculate the sum of all multipl.results for that frequency
  sumVal = 0.0;
  for (unsigned n=fromXp; n < toXp; n++)
  {
    sumVal = sumVal + pLocalResults[n - fromXp];
  }
  // Divide th result by the num. of samples to get an average value
  ret = sumVal / (double)(toXp - fromMs);
  //
  zurueck:
  //if (this->verbose)
  //    ::printf("WAVE::processDFT(3args) freeing memory pLocalResults\n"); 
  if (pLocalResults != NULL)
    delete pLocalResults;
  if (this->verbose)
  {
    ::printf("WAVE::processDFT(3args) ret=%lf\n", ret);
    ::fflush(stdout);
  }
  return ret;
}   // end WAVE::processDFT()

// ----------------------------------------------------------
// Process the inverse Discrete Fourier Transform, i.e. a synthesis
// Params: noHertz =base freuency (= harmonic 1)
//         maxSamples: number of samples to be returned (MONO only)
//         harmonics = array of numbers for the harmonics 
//         numHarmonics = number of entries in array harmonics
// Returns: an allocated array of results (the synthesized WAVE samples
//          may return NULL!
double * WAVE::processInvDFT(unsigned  noHertz,
                              unsigned long maxSamples,
                              unsigned *harmonics, // ex 1,3,5,7
                              unsigned numHarmonics)
{
  int ret = 0;
  double *pResults = NULL;
  WAVE   *pHarm = NULL;
  int    amplitude = 40;
  int    loopCount = 1;
  double *pNew;
  char   fn[196];
  //
  if (this->verbose)
    ::printf("WAVE::processInvDFT() noHerz=%u maxSaples=%lu\n", 
              noHertz, maxSamples);
  pResults = new double[maxSamples];
  if (pResults == NULL)
  {
    ret = -1;
    goto zurueck;
  }
  // calculate the base sine wave (=1st harmonic)
  // This results in normalized data (-1 : +1)
  this->makeWaveFormSine(noHertz, maxSamples);
  this->changeAmplitude(amplitude);
  //
  // generate now overtone sine waves (harmonics)
  // Add them to the base sine wave
  amplitude = 35; // in percent
  for (unsigned nh = 0; nh < numHarmonics; nh++)
  {
    if (this->verbose)
      ::printf("WAVE::processInvDFT() no. harmonic=%u\n", 
                nh);
    if (harmonics[nh] == 1)
    {
      // this is the base wave  - ignore it (has been calculated already)
      if (this->verbose)
        ::printf("WAVE::processInvDFT() for 1st harmonic: nothing to do!\n");
      continue;
    }
    //
    // calculate the amplitudes for the base wave (=1st harmonic)
    pHarm = new WAVE(this->samplesSec,
                      maxSamples / this->samplesSec,
                      1);  // mono!
    // this will produce normalized samples (-1:+1)
    pHarm->makeWaveFormSine(noHertz * harmonics[nh], 
                            maxSamples);
    //pHarm->protocolDataPortion(256L, "portion A");
    //::sprintf(fn, "HarmonicA%01u.wav", nh);
    //pHarm->write(fn);
    //
    pHarm->changeAmplitude(amplitude);
    //pHarm->protocolDataPortion(256L, "portion B");
    // std:to:string() does not work - so this version!
    //::sprintf(fn, "HarmonicB%01u.wav", nh);
    //pHarm->write(fn);
    //
    // Add the amplitudes for the harmonic to the base wave
    // The base wave is retained with 50 percent
    // The following overtones are added with decreased precentages
    pNew = pHarm->getInpData();
    for (unsigned long n=0; n < maxSamples; n++)
    {
      this->pInpData[n] = this->pInpData[n] + pNew[n];
    }
    delete pHarm;
    amplitude -= 15;
    if (amplitude < 0)
      amplitude = 10;
  }   // end for nh = 0 ...
  //
  // the new wave is now in the this instance
  //
  zurueck:
  //
  if (this->verbose)
  {
    ::printf("WAVE::processInvDFT() ret=%d\n", ret);
    ::fflush(stdout);
  }
  //
  return pResults;
}   /// end WAVE::processInvDFT()

// ---------------------------------------------------------------
// This is needed for some input tools like Gnuplot etc.
// Output all data pairs (x,y) to stdout
void WAVE::outputData(const char *pMsg,
                      unsigned processMaxSamples)
{
  ::printf("%s\n", pMsg);
  for (unsigned i=0; i < processMaxSamples; i++)
  {
    ::printf("%05u %6.2lf\n", i,  this->pInpData[i]);
  }
}   // end WAVE:outputData()

// --------------------------------------------------------------
// Allocates array pResults and keeps the results in it
// for later evaluation
void WAVE::keepResult(double result)
{
  if (this->pResults == NULL)
  {
    this->maxResults = (this->freqEnd - this->freqStart) / this->freqStep;    
    this->pResults = new double[this->maxResults];
    this->resCount = 0;
  }
  if (resCount < this->maxResults)
    this->pResults[this->resCount++] = result;
}   // end WAVE::keepResult()

// --------------------------------------------------------------
// Assumes that keepResult() has been called before
// Param. onlyOne: if 0 then all harmonics are output
//               : if 1 then only the strongest harmonic is output
void WAVE::findResults(int onlyOne)
{  
  double   bestResult  = 0.0;
  unsigned bestFreq  = 0;
  unsigned lastBestFreq = 0;
  //
  for (unsigned nHarm = 1; nHarm < 5; nHarm++)
  {
    bestResult = 0.0;
    bestFreq   = 0;
    int i      = 0;
    int lastI  = -1;
    for (unsigned n=this->freqStart; n <= this->freqEnd; n += this->freqStep)
    {
      if (i >= this->maxResults)
        continue;
      if (this->pResults[i] >= bestResult)
      {
        bestResult = this->pResults[i];
        bestFreq   = n;
        lastI      = i;
      }
      i++;
    }   // end for ...
    if (onlyOne)
    { 
      if (bestFreq < lastBestFreq)
        break;
    }
    ::printf("WAVE::findResults() Harmonic=%u Result=%lf frequency=%u\n",
              nHarm, bestResult, bestFreq);
    if ((lastI >= 0) && (lastI < this->maxResults))
      this->pResults[lastI] = 0.0;
    lastBestFreq = bestFreq;
  }   // end for nHarm ...
}   // end find WAVE::findResults()

// ------------------------------------------------------------------
// Schreibt den Dateikopf mit dem Dateityp "RIFF"
// Returns: the number of bytes written
long WAVE::writeMainChunk(int ofd)
{
  int  nb;
  char buf[20];
  MAIN_CHUNK header;
  //
  ::memcpy((char *)&header.main_chunk, "RIFF", 4);
  header.mlength = 36 + 
                   (this->noSamples *
                    this->noChannels *
                    this->blockAlign);
  ::memcpy((char *)&header.chunk_type, "WAVE", 4);
  nb = ::write(ofd, 
               (char *)&header, 
                sizeof(header));
  //
  if (this->verbose)
    ::printf("WAVE::writeMainChunk() ret=%d\n", nb);
  return nb;
}   // end WAVE::writeMainChunk()

// --------------------------------------------------------------------
// Returns: the number of bytes(!) written
// Hinweis: offensichtlich darf nur ein data-frame erzeugt werden!!!
// Alle Versuche, es anders zu machen (z.B. 1 frame pro Sekunde) schlugen fehl!
// Returns: the number of bytes written (not frames!)
//          < 0L upon error
long WAVE::writeDataChunk(int ofd,
                          unsigned bitsPerSample)// may override this->bitsSample
{
  long     noWritten     = 0L;
  DATA_CHUNK header;
  unsigned sampleSize    = 2;
  long     nb            = 0L;
  long     noData        = 0L;
  short    *pShortData   = NULL;
  unsigned long shortPos = 0L;
  unsigned long bytePos  = 0L;
  unsigned char *pByteData = NULL;
  //
  if (this->verbose)
  {
    ::printf("WAVE::writeDataChunk() start\n");
    ::fflush(stdout);
  }
  sampleSize = sizeof(short);
  if (bitsPerSample < 16)
    sampleSize = sizeof(char);
  //
  // Header fuer den Frame schreiben (8 Bytes)
  ::strncpy((char *)&header.data_type, "data", 4);
  // TODO: blockAlign missing in formula;
  header.dlength = this->getNoAllocatedSamples() * sampleSize;
  // blockAlign here added
  header.dlength += (this->noChannels * bitsPerSample) / 8;
  header.dlength += this->blockAlign;
  nb = ::write(ofd, (char *)&header, sizeof(header));
  if (nb <= 0)
  {
    noWritten = -1L;
    goto zurueck;
  }
  noWritten += nb;
  noData    = this->getNoAllocatedSamples();
  //
  if (sampleSize == sizeof(short))
  {
    // short sample data requested 
    pShortData = this->convertSamplesToShort();
    shortPos   = 0L;
    if (this->verbose)
    {
      ::printf("WAVE::writeDataChunk() no.of samples=%lu sampleSize=%u\n", 
                noData, sampleSize);
      ::fflush(stdout);
    }
    while (noData > 0L)
    {
      nb       = PORTION_SIZE * sizeof(short);
      if (nb > (noData * sizeof(short)))
        nb = noData * sizeof(short);
      nb       = ::write(ofd, (char *)&pShortData[shortPos], (int)nb);
      if (nb < 0)
      {
        // EOF or error
        if (this->verbose)
          ::printf("WAVE::writeDataChunk() ERROR after %lu bytes\n",
                    noWritten);
        break; 
      }
      noData    -= (nb / sizeof(short));
      shortPos  += (nb / sizeof(short));
      noWritten += nb;    
    }   // end while ...
  }   // end if (sampleSize == sizeof(short))
  else
  {
    // 8-bit samples requested
    pByteData = this->convertSamplesToByte();
    bytePos   = 0L;
    while (noData > 0L)
    {
      nb       = PORTION_SIZE;
      if (nb > noData)
        nb = noData;
      nb       = ::write(ofd, (char *)&pByteData[bytePos], (int)nb);
      if (nb < 0)
      {
        // EOF or error
        if (this->verbose)
          ::printf("WAVE::writeDataChunk() ERROR after %lu bytes\n",
                    noWritten);
        break; 
      }
      noData    -= nb;
      bytePos   += nb;
      noWritten += nb;    
    }   // end while ...
  }   // end else
  //
  // TODO: not sure if the blockAlign bytes must be written here!
  //
  zurueck:
  if (pShortData != NULL)
    delete pShortData;
  if (pByteData != NULL)
    delete pByteData;
  if (this->verbose)
    ::printf("WAVE::writeDataChunk() ret=%ld\n", noWritten);
  return noWritten;
}   // end WAVE::writeDataChunk()

// --------------------------------------------------------------------
// Returns: the number of bytes written
//          <= 0 upon error
long WAVE::writeFormatChunk(int ofd,
                            unsigned bitsPerSample)
{
  long nb = 0L;
  char buf[20];
  SUB_CHUNK header;
  //
  ::memcpy((char *)&header.sub_chunk, "fmt ", 4);
  header.slength   = 16;
  header.format    = 1;
  header.channels  = this->noChannels;       
  header.sample_fq = this->samplesSec;      
  header.bytes_per_second = this->bytesSec;
  header.block_align      = this->noChannels * ((bitsPerSample + 7) / 8);
  header.bits_per_sample  = bitsPerSample; 
  nb = ::write(ofd, (char *)&header, sizeof(header));
  //
  if (this->verbose)
    ::printf("WAVE::writeFormatChunk(): ret=%ld\n", nb);
  return nb;
}   // end WAVE::writeFormatChunk()

// ---------------------------------------------------------------
// Writes a WAVE file with the given bits-per-sample value
// Note: denormalization (mapping from 1.0 .. -1.0 to short or byte
//       values is no longer done here
// Returns: >= 0L if  OK
//          < 0L upon error
long WAVE::write(const char *outFn,
                 unsigned bitsPerSample)
{
  long  lRet = 0L;
  int   ofd = -1;
  long  nb  = 0L;
  unsigned long writePos = 0L;
  //
  if (this->verbose)
    ::printf("WAVE::write(): start bitsPerSample=%u\n", 
              bitsPerSample);
#ifdef WIN32
  //_set_fmode(_O_BINARY);  // sets global binary mode for IO
  ofd = ::creat(outFn, O_WRONLY | S_IWRITE);
  _setmode(ofd, _O_BINARY);  // _setmode() instead of setmode()!
#else
  ofd = ::creat(outFn, O_WRONLY | S_IRWXU);
#endif
  if (ofd < 0)
  {
    lRet = -1L;
    printf("WAVE::write(): ERROR: cannot open output file\n");
    goto zurueck;
  }
  if (this->verbose)
    ::printf("WAVE::write(): output file=%s is open\n", outFn);
  //
  // write the wave file header
  nb = this->writeMainChunk(ofd);
  if (nb < 0)
  {
    lRet = -2L;
    ::printf("WAVE::write(): ERROR: write error output file\n");
    goto zurueck;
  }
  writePos += nb;
  if (this->verbose)
    ::printf("WAVE::write(): after main chunk write pos=%ld\n", 
              writePos);
  //
  // write the header for the data chunk
  nb = this->writeFormatChunk(ofd, bitsPerSample);
  if (this->verbose)
    ::printf("WAVE::write(): writeFormatChunk() ret=%ld\n", nb);
  if (nb < 0)
  {
    if (this->verbose)
      ::printf("WAVE::write() ERROR writing the format chunk\n");
    lRet = -3L;
    goto zurueck;
  }
  writePos += nb;
  if (this->verbose)
    ::printf("WAVE::write(): after format chunk write pos=%ld\n", writePos);
  //
  // write the data (samples)
  nb = this->writeDataChunk(ofd,
                            bitsPerSample);
  if (nb < 0)
  {
    if (this->verbose)
      ::printf("WAVE::write() ERROR writing the data chunk\n");
    lRet = -4L;
    goto zurueck;
  }
  if (this->verbose)
    ::printf("WAVE::write(): data chunk written OK\n"); 
  writePos += nb;
  //
  lRet = writePos;  // means OK
  //
  zurueck:
  if (ofd > -1)
    ::close(ofd);
  if (this->verbose)
    ::printf("WAVE::write(): ret=%ld\n", lRet);
  return lRet;
}   // end WAVE::write()

// ---------------------------------------------------------------
// Writes a WAVE file with the sample resolution as given in 
// the WAVE instance
// Returns: >= 0L if  OK
//          < 0L upon error
long WAVE::write(const char *outFn)
{
  long  lRet = 0L;
  //
  lRet = this->write(outFn, this->getBitsSample());
  //
  if (this->verbose)
    ::printf("WAVE::write(): ret=%ld\n", lRet);
  return lRet;
}   // end WAVE::write()

// ---------------------------------------------------------------
// Writes an ASCII file containing per line x corrdinate and value
// This file can be read back by GNUPLOT etc.
// Returns: >= 0L if  OK
//          < 0L upon error
long WAVE::writeASCII(const char *outFn,
                      unsigned long noSamp)
{
  long  lRet = 0L;
  FILE *ofd = NULL;
  long  nb  = 0L;
  unsigned long writePos = 0L;
  //
  ofd = ::fopen(outFn, "w");
  if (ofd == NULL)
  {
    lRet = -1L;
    printf("WAVE::writeASCII(): ERROR: cannot open output file\n");
    goto zurueck;
  }
  if (this->verbose)
    ::printf("WAVE::writeASCII(): output file=%s is open\n", outFn);
  //
  // write the data (samples)
  for (unsigned long n=0; n < noSamp; n++)
  {
    nb = ::fprintf(ofd, "%08lu %lf\n", n, this->pInpData[n]);
    if (nb < 0)
    {
      lRet = -2L;
      goto zurueck;
    }
    writePos += 1;
  }
  //
  lRet = writePos;  // means OK
  //
  zurueck:
  if (ofd == NULL)
    ::fclose(ofd);
  if (this->verbose)
    ::printf("WAVE::writeASCII(): ret=%ld\n", lRet);
  return lRet;
}   // end WAVE::writeASCII()

// ------------------------------------------------------------
// Returns an allocated buffer with short data
// No normalization or denormalization done here!
// The returned buffer must be freed externally
short * WAVE::convertSamplesToShort()
{
  short *pShortData = NULL;
  unsigned long noShorts;
  double dValue;
  //
  if (this->verbose)
  {
    ::printf("WAVE::convertSamplesToShort(): start\n");
    ::fflush(stdout);
  }
  noShorts   = this->getNoAllocatedSamples();
  pShortData = new short[noShorts];
  if (this->verbose)
  {
    ::printf("WAVE::convertSamplesToShort(): noShorts=%lu\n", 
               noShorts);
    ::fflush(stdout);
  }
  for (unsigned long i=0; i < noShorts; i++)
  {
    dValue = this->pInpData[i];
    // avoid overflow in conversion to short
    dValue = dValue / 1.15;     
    pShortData[i] = (short)dValue;
  }
  if (this->verbose)
  {
    ::printf("WAVE::convertSamplesToShort(): done\n");
    ::fflush(stdout);
  }
  return pShortData;
}  // end WAVE::convertSamplesToShort()

// ------------------------------------------------------------
// Returns an allocated buffer with unsigned byte data
// No normalization or denormalization done here!
// The returned  buffer must be freed externally
unsigned char * WAVE::convertSamplesToByte()
{
  unsigned char  *pByteData = NULL;
  unsigned long  noBytes;
  short          sValue;
  unsigned char  ucValue;
  //
  if (this->verbose)
  {
    ::printf("WAVE::convertSamplesToByte(): start\n");
    ::fflush(stdout);
  }
  noBytes   = this->getNoAllocatedSamples();
  pByteData = new unsigned char[noBytes];
  if (this->verbose)
  {
    ::printf("WAVE::convertSamplesToByte(): noBytes=%lu\n", 
               noBytes);
    ::fflush(stdout);
  }
  for (unsigned long i=0; i < noBytes; i++)
  {
    sValue = (short)(this->pInpData[i]);
    ucValue = ((long)sValue + 32767) >> 8;
    pByteData[i] = ucValue;
  }   // end for ...
  //
  if (this->verbose)
  {
    ::printf("WAVE::convertSamplesToByte(): done\n");
    ::fflush(stdout);
  }
  return pByteData;
}  // end WAVE::convertSamplesToByte()

// --------------------------------------------------------------
void resetSignal()
{
  ::pthread_mutex_lock(&::mutex1);
  ::eventMask = 0; 
  ::pthread_mutex_unlock(&mutex1);
}   // end resetSignal()

// ----------------------------------------------------
#ifdef WIN32
static WAVEHDR      waveHdr1;
static WAVEHDR      waveHdr2;

// ----------------------------------------------------
// MIcrosoft doku says that we should not use any waveIn..() functions
// in this callback routine...!
long CALLBACK waveInCallback(HWAVEIN    hwi,
                             UINT       message, 
                             DWORD_PTR  param1,  // pointer to WAVE class instance
                             DWORD_PTR  param2,  // pointer waveHdr
                             DWORD_PTR  param3)
{
  static unsigned long noShortsRecorded = 0L;
  unsigned noShorts = 0;
  PWAVEHDR pWaveHdr = (PWAVEHDR)param2;
  WAVE *pWave       = (WAVE *)param1;
  //
  if (::verbose)
    ::printf("waveInCallback() %u\n", message);
  //
  switch (message)
  {
    case MM_WIM_DATA:
      if (::verbose)
        ::printf("waveInCallback() MM_WIM_DATA %lu bytes\n",
                   pWaveHdr->dwBytesRecorded);
      noShorts = pWaveHdr->dwBytesRecorded / sizeof(short);
      noShortsRecorded += noShorts;
      if (noShortsRecorded >= pWave->getNoAllocatedSamples())
      {
        // overflow of data - stop
        // set the signal saying  "recording stops"
        ::pthread_mutex_lock(&::mutex1);
        ::eventMask |= 1;
        ::pthread_mutex_unlock(&::mutex1);
        if (::verbose)
          ::printf("waveInCallback() OVERFLOW signal set\n");
      }
      else
      {
        // continue recording
        if (::verbose)
          ::printf("waveInCallback() continue recording 1\n");
        pWave->addSamples((short *)(pWaveHdr->lpData), 
                          noShorts);
        // as we cannot use here waveInAddBuffer() we set a signal instead
        ::pthread_mutex_lock(&::mutex1);
        if (pWaveHdr == &::waveHdr1)
          ::eventMask |= 2;
        else
          ::eventMask |= 4;
        ::pthread_mutex_unlock(&::mutex1);
        if (::verbose)
          ::printf("waveInCallback() continue recording 2\n");
      }
      break;
    case MM_WIM_OPEN:
      if (::verbose)
        ::printf("waveInCallback() MM_WIM_OPEN\n");
      noShortsRecorded = 0L;
      pWave->setNoSamples(0L);
      break;
    case MM_WIM_CLOSE:
      // never got that event!
      if (::verbose)
        ::printf("waveInCallback() MM_WIM_CLOSE\n");
      break;
    default:
      if (::verbose)
        ::printf("waveInCallback() default\n");
      break;
  }   // end switch
  //
  if (::verbose)
    ::printf("waveInCallback() returns\n");
  return 0L;
}   // end waveInCallback()

// ---------------------------------------------------------------
// Prepares an input buffer 
// Call also waveInAddBuffer()
long prepareInputBuffer(HWAVEIN  hWaveIn,
                        PWAVEHDR pWaveHdr,
                        char     *pBuffer,
                        unsigned bufferSize)
{
  long         lRet = 0L;
  MMRESULT     mmResult = 0;
  //
  if (::verbose)
    ::printf("prepareInputBuffer() start\n");
  // Set up two  headers and prepare tem
  pWaveHdr->lpData          = (char *)(pBuffer);
  pWaveHdr->dwBufferLength  = bufferSize ;
  pWaveHdr->dwBytesRecorded = 0L;
  pWaveHdr->dwUser          = 0;
  pWaveHdr->dwFlags         = 0;
  pWaveHdr->dwLoops         = 1;
  pWaveHdr->lpNext          = NULL;
  pWaveHdr->reserved        = 0;
  //
  mmResult = ::waveInPrepareHeader(hWaveIn, 
                                   pWaveHdr, 
                                   sizeof(WAVEHDR));
  //
  ::waveInAddBuffer (hWaveIn, 
                     pWaveHdr, 
                     sizeof(WAVEHDR));
  //
  zurueck:
  if (::verbose)
    ::printf("prepareInputBuffer() lRet=%ld\n", lRet);
  ::fflush(stdout);
  return lRet;
}   // end prepareInputBuffer()

// -----------------------------------------------------------
// Records audio data
// Returns: the number of samples recorded
// Returns: < 0 if error > 0 = no. of samples recorded
// Note: waveInOpen() reports an error if no. of channels > 1!!
long WAVE::record(unsigned long durMs)  // WIN32 version!
{
  long lRet  = 0;
  int          cRet = 0;
  static HWAVEIN      hWaveIn;
  static WAVEFORMATEX waveform;
  MMRESULT            mmResult;
  static char *       pBuffer1 = NULL;
  static char *       pBuffer2 = NULL;
  unsigned            bufferSize = 8192;
  unsigned long noFrames = 0L;
  unsigned            localEventMask = 0;
  bool                success  = true;
  unsigned int        noDevs   = 0;
  WAVEINCAPS          caps;
  //
  noDevs =  ::waveInGetNumDevs();
  if (this->verbose)
  {
    ::printf("WAVE::record() start no.hannels=%u samplesSec=%u no.inp.devices=%u\n",
            this->noChannels, this->samplesSec, noDevs);
    for (unsigned int i=0; i < noDevs; i++)
    {
      mmResult = ::waveInGetDevCaps(i, &caps, sizeof(caps));
      ::printf("WAVE::record() capabilities dev no=%u name=%s channels=%u formats=%0x08\n",
                i, caps.szPname, (WORD)caps.wChannels, (DWORD)caps.dwFormats);
    }
  }
  //
  // initialize the signal for keyboard interrupt
  // This same signal is also used to indicate the reception of a buffer
  ::resetSignal();
  //
  noFrames = (long)(this->samplesSec * (durMs / 1000L));
  this->allocateSampleData((unsigned long)(noFrames * this->noChannels));
  //
  // Open waveform audio for input - this returns a handle in hWaveIn
  // Note: we try here 2 times as waveInOpen() sometimes reacts
  //       with a "bad format" error if noChannels > 1
  success = false;
  for (int noTrial=0; noTrial < 2; noTrial++)  // try two times
  {
    waveform.wFormatTag      = this->format;
    waveform.nChannels       = this->noChannels;
    waveform.nSamplesPerSec  = this->samplesSec;
    waveform.nAvgBytesPerSec = this->bytesSec;
    waveform.nBlockAlign     = this->blockAlign;
    waveform.wBitsPerSample  = this->bitsSample;
    waveform.cbSize          = 0;
    if (this->verbose)
      ::printf("WAVE::record() starting waveInOpen()\n");
    ::fflush(stdout);
    mmResult = ::waveInOpen(&hWaveIn, 
                            WAVE_MAPPER, 
                            &waveform,
                            (DWORD)waveInCallback, // callback function  
                            (DWORD)this, 
                            CALLBACK_FUNCTION);  // 
    if (this->verbose)
    {
      ::printf("WAVE::record() waveInOpen() mmResult=%ld\n", 
            (long)mmResult);
      ::fflush(stdout);
    }
    if (mmResult == MMSYSERR_NOERROR)
    {
      success = true;
      break;
    }
    // error here - probably due to the fact that stereo is not supported
    else
    {
      char aux[256];
      ::waveInGetErrorText(mmResult, aux, sizeof(aux));
      if (this->verbose)
        ::printf("WAVE::record() error from waveInOpen()=%s\n", aux);
                  //::reportLastError(mmResult));
      // try with mono instead of stereo
      this->noChannels = 1;      // fix 30.08.2021
      if (this->verbose)
        ::printf("WAVE::record() trying with mono instead of stereo\n");
      // try again ...
    }
  }   // end for noTrial ...
  if (! success)
  {
    lRet = -1L;
    goto zurueck;
  }
  // 
  // allocate buffer for the data to be received
  // Note: this allocates a double array in the WAVE class
  // The Windows API calls operate with bytes and need separate buffers
  if (this->verbose)
    ::printf("WAVE::record() preparing allocating buffer\n");
  ::fflush(stdout);
  pBuffer1 = new char[bufferSize];
  lRet = ::prepareInputBuffer(hWaveIn,
                              &waveHdr1,  
                              pBuffer1, 
                              bufferSize);
  if (lRet < 0)
  {
    lRet = -2L;
    goto zurueck;
  }
  //
  pBuffer2 = new char[bufferSize];
  lRet = ::prepareInputBuffer(hWaveIn,
                              &waveHdr2,  
                              pBuffer2, 
                              bufferSize);
  if (lRet < 0)
  {
    lRet = -3L;
    goto zurueck;
  }
  ::fflush(stdout);
  //
  // start recording
  if (this->verbose)
    ::printf("WAVE::record() starting the recording\n");
  mmResult = ::waveInStart(hWaveIn);
  if (mmResult != MMSYSERR_NOERROR)
  {
    lRet = -4L;
    goto zurueck;
  }
  if (this->verbose)
    ::printf("WAVE::record() 7 mmResult=%ld\n", (long)mmResult);
  //
  // check an interrupt signal (keyboard!) while recording
  while (durMs > 0)
  {
    // interrupt (via keyboard) signalled?
    // overflow signalled?
    int rc = ::pthread_mutex_lock(&::mutex1);
    if (rc != 0)
    {
       // cannot lock the mutex
       if (this->verbose)
         ::printf("WAVE::record() ERROR canot lock mutex\n");
       goto goOn;
    }
    localEventMask = ::eventMask;
    ::pthread_mutex_unlock(&mutex1);
    if (localEventMask & 1)
    {
      if (this->verbose)
        ::printf("WAVE::record() INTERRUPT detected - we stop recording\n");
      break;  // leave this loop
    }
    if (localEventMask & 2)
    {
      if (this->verbose)
        ::printf("WAVE::record() signal buffer 1 received\n");
      ::resetSignal();
      mmResult = ::waveInUnprepareHeader(hWaveIn, 
                                         &waveHdr1, 
                                         sizeof(WAVEHDR));
      delete pBuffer1;
      pBuffer1 = new char[bufferSize];
      mmResult = ::prepareInputBuffer(hWaveIn,
                              &waveHdr1,  
                              pBuffer1, 
                              bufferSize);
      if (this->verbose)
        ::printf("WAVE::record() buffer 1 added again\n");
    }
    if (localEventMask & 4)
    {
      if (this->verbose)
        ::printf("WAVE::record() signal buffer 2 received\n");
      ::resetSignal();
      mmResult = ::waveInUnprepareHeader(hWaveIn, 
                                         &waveHdr2, 
                                         sizeof(WAVEHDR));
      delete pBuffer2;
      pBuffer2 = new char[bufferSize];
      mmResult = ::prepareInputBuffer(hWaveIn,
                              &waveHdr2,  
                              pBuffer2, 
                              bufferSize);
      if (this->verbose)
        ::printf("WAVE::record() buffer 2 added again\n");;
    }
    //
  goOn:
    durMs--;
    ::Sleep(1L);
  }   // end while durMs ...
  if (this->verbose)
    ::printf("WAVE::record() recording loop left no samples=%lu\n", 
               this->getNoSamples());
  //
  mmResult = ::waveInReset(hWaveIn);
  if (this->verbose)
    ::printf("WAVE::record() waveInReset() OK\n");
  ::waveInStop(hWaveIn);
  if (this->verbose)
    ::printf("WAVE::record() waveInStop() OK)\n");
  ::waveInClose(hWaveIn);
  if (this->verbose)
    ::printf("WAVE::record()waveInClose() OK\n");
  //
  // everything is OK here
  lRet = this->getNoSamples();
  //
  zurueck:
  if (pBuffer1 != NULL)
    delete pBuffer1;
  if (pBuffer2 != NULL)
    delete pBuffer2;
  if (this->verbose)
    ::printf("WAVE::record() lRet=%ld\n", lRet);
  //
  return lRet;
}   // end WAVE::record() WIN32 version

// -----------------------------------------------------------------
// WIN32 version!
// Note: we expect samples (data) in internal double format
//       These data will always be converted to signed short data before play
int WAVE::play(HANDLE terminationEvent,
               unsigned long startMs,
               unsigned long durationMs)
{
  int      ret = 0;
  int      cRet;
  MMRESULT mmRet;
  DWORD    dwRet;
  DWORD    dword;
  DWORD    dwRepetitions = 1; 
  double   noMs = 0.0;
  PWAVEHDR pWaveHdr = NULL;
  HWAVEOUT hWaveOut = 0;
  HANDLE   evHandle;
  static WAVEFORMATEX  waveForm;
  short    *pShortData = NULL;
  //
  if (this->verbose)
    ::printf("WAVE::play() start filename=%s samplesSec=%lu channels=%u dur=%lu\n", 
               this->filename, this->samplesSec, this->noChannels,
               durationMs);
  pWaveHdr = (PWAVEHDR)new char[sizeof(WAVEHDR)];
  //pWaveHdr->lpData          = (char *)(this->pInpData);
  //
  // we must convert the internal double values(samples) to shorts !
  if (this->verbose)
  {
    ::printf("WAVE::play() converting %lu double samples to short\n",
              this->getNoAllocatedSamples());
    ::fflush(stdout);
  }
  pShortData = this->convertSamplesToShort();
  if (this->verbose)
  {
    ::printf("WAVE::play() conversion OK\n");
    ::fflush(stdout);
  }
  //
  pWaveHdr->lpData          = (char *)pShortData;
  pWaveHdr->dwBufferLength  = this->getNoAllocatedSamples() * sizeof(short);
  pWaveHdr->dwBytesRecorded = 0;
  pWaveHdr->dwUser          = 0;
  // if only one buffer ist used both flags must be set
  pWaveHdr->dwFlags         = WHDR_BEGINLOOP |
                              WHDR_ENDLOOP;
  pWaveHdr->dwLoops         = dwRepetitions;
  pWaveHdr->lpNext          = NULL;
  pWaveHdr->reserved        = 0;
  //
  waveForm.wFormatTag      = this->format;         // normally 1=WAVE_FORMAT_PCM
  waveForm.nChannels       = this->noChannels;   
  waveForm.nSamplesPerSec  = this->samplesSec;  
  waveForm.nAvgBytesPerSec = this->samplesSec * sizeof(short) * this->noChannels; 
  waveForm.wBitsPerSample  = sizeof(short) * 8;   //this->bitsSample;
  // block align: values 1,2(16 bit mono) or 4(16 bit stereo)
  // from: https://docs.microsoft.com/en-us/windows/win32/api/mmeapi/ns-mmeapi-waveformatex  
  // "nBlockAlign must be equal to the product of nChannels and wBitsPerSample divided by 8"
  // waveForm.nBlockAlign     = sizeof(short) * this->noChannels;
  waveForm.nBlockAlign     = (this->noChannels * waveForm.wBitsPerSample) / 8;
  if (this->verbose)
  {
    ::printf("WAVE::play() format=%u channels=%lu samplesSec=%lu bitsSamples=%u\n",
              waveForm.wFormatTag, 
              waveForm.nChannels, 
              waveForm.nSamplesPerSec,
              waveForm.wBitsPerSample);
    ::fflush(stdout);
  }
  //
  // Note: the audio device must be opend for each WAVE file (each thread)
  if (hWaveOut == 0)
  {
    evHandle = ::CreateEvent(0, false, false, 0);
    mmRet    = ::waveOutOpen(&hWaveOut, 
                             WAVE_MAPPER, 
                             (WAVEFORMATEX *)(&waveForm),
                             (unsigned long)(evHandle), 
                             0L, 
                             CALLBACK_EVENT); 
    if (mmRet != MMSYSERR_NOERROR)
    {
      char msg[256];
      ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
      ::printf("WAVE::play() waveOutOpen() reports ERROR: %s\n", msg);
      ret = -1;
      goto zurueck;
    }
    if (this->verbose)
      printf("WAVE::play() waveOutOpen() finished\n");
    // not sure whether this is needed 
    mmRet = ::waveOutReset(hWaveOut);
    if (mmRet != MMSYSERR_NOERROR)
    {
      char msg[128];
      ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
      ::printf("WAVE::play() waveOutReset() reports ERROR: %s\n", msg);
      ret = -2;
      goto zurueck;
    }
    if (this->verbose)
      ::printf("WAVE::play() waveOutReset() finished\n");
    ret = 1;
  }  // if (hWaveOut == 0)
  //
  mmRet = ::waveOutPrepareHeader(hWaveOut, 
                                 pWaveHdr, 
                                 sizeof(WAVEHDR));
  if (mmRet != MMSYSERR_NOERROR)
  {
    char msg[256];
    ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
    ::printf("WAVE::play() waveOutPrepareHeader() reports ERROR: %s\n", msg);
    ret = -4;
    goto zurueck;
  }
  //
  mmRet = ::waveOutWrite(hWaveOut, 
                         pWaveHdr, 
                         sizeof(WAVEHDR));
  if (mmRet != MMSYSERR_NOERROR)
  {
    char msg[256];
    ::waveOutGetErrorText(mmRet, msg, sizeof(msg) - 1);
    ::printf("WAVE::play() waveOutWrite() reports ERROR: %s\n", msg);
    ret = -5;
    goto zurueck;
  }
  //
  // Wait for events until we detect WHDR_DONE in the WAVEHDR structure
  // We sent then a termination event to the main process
  while (1)
  {
    dwRet = ::WaitForSingleObject(evHandle, (DWORD)10);
    if (this->verbose > 1)
     ::printf("WAVE::play() event received ret=%ld\n", (long)dwRet);
    if ((pWaveHdr->dwFlags) & WHDR_DONE)
    {
      if (this->verbose)
        ::printf("WAVE::play() WHDR_DONE detected\n");
      // send the termination event to the main process
      if (terminationEvent != 0)
        ::SetEvent(terminationEvent);
      break;
    }
     // no Sleep() needed!
  }   // end while (1)
  //
  ret = 0;
  //
  zurueck:
  if (pShortData != NULL)
    delete pShortData;
  if (pWaveHdr != NULL)
    delete pWaveHdr;
  ::waveOutClose(hWaveOut);
  if (this->verbose)
    ::printf("WAVE::play() ret=%d\n", ret);
  return ret;
}   // end WAVE::play()  WIN32 version

#else

// Linux/Alsa here
#include <time.h>
#include <signal.h>

#define BUFSIZE 512
#define RATE    44100

// ---------------------------------------------------------
// Handles CTRL-C
// TODO: not finished!
void sigint(int sig) 
{
/*
  if (fout != NULL) 
	{
    fclose(fout);
  }
  exit(1);
*/
}   // end sigint()


// ----------------------------------------------------------
// Record wave data from the standard device
// The data (samples) in pInpData are destroyed/replaced
// Returns: < 0 upon error
// Note: assumes that pAlsa->open() has been run
long WAVE::record(ALSA *pAlsa,
                  unsigned long durMs)
{
  long lRet = 0L;
	int i;
  int err;
  static short buf[BUFSIZE];   // must be static and not longer than 512
  snd_pcm_t *capture_handle;
  snd_pcm_hw_params_t *hw_params;	
  int nread;
  snd_pcm_format_t format;
	unsigned int rate; //snd_pcm_format_t rate;
  unsigned count = 0;
  unsigned long maxNoData = 0L;
  unsigned int localEventMask = 0;
  //
  if (this->verbose)
    printf("WAVE::record() no.channels=%u durMs=%lu samplesSec=%lu\n", 
            this->noChannels, durMs, this->samplesSec);
	if (pAlsa == NULL)
  {
	  lRet = -100L;
		goto zurueck;
	}
  capture_handle = pAlsa->getHandle();
  //
  // initialize the signal for keyboard interrupt
  // This same signal is also used to indicate the reception of a buffer
  ::resetSignal();
	//
	::signal(SIGINT, ::sigint);
	//	   
  err = ::snd_pcm_hw_params_malloc(&hw_params);
	if (err < 0)
  {
	  lRet = -2L;
		goto zurueck;
	}
				 
  err = ::snd_pcm_hw_params_any(capture_handle, 
                                hw_params);
	if (err < 0)
  {
	  lRet = -3L;
		goto zurueck;
	}
	
  err = ::snd_pcm_hw_params_set_access(capture_handle, 
                                       hw_params, 
																			 SND_PCM_ACCESS_RW_INTERLEAVED);
 	if (err < 0)
  {
	  lRet = -4L;
		goto zurueck;
	}

  err = ::snd_pcm_hw_params_set_format(capture_handle, 
                                       hw_params, 
																			 SND_PCM_FORMAT_S16_LE);
	if (err < 0)
  {
	  lRet = -5L;
		goto zurueck;
	}
	
  rate = (this->samplesSec);
  err = ::snd_pcm_hw_params_set_rate_near(capture_handle, 
                                          hw_params, 
                                          &rate, 0);
	if (err < 0)
  {
	  lRet = -6L;
		goto zurueck;
	}
	
  err = ::snd_pcm_hw_params_set_channels(capture_handle, 
                                         hw_params, 
                                         2);
	if (err < 0)
  {
	  lRet = -7L;
		goto zurueck;
	}
	
  err = ::snd_pcm_hw_params(capture_handle, 
                            hw_params);
	if (err < 0)
  {
	  lRet = -8L;
		goto zurueck;
	}
	
  ::snd_pcm_hw_params_free(hw_params);
  //
  maxNoData = (this->samplesSec * this->noChannels) * 10L;
  this->allocateSampleData(maxNoData);  // 10 seconds
	//
  //while (1) // EH: muss mit CTL-C beendet werden!
  //while (lRet < 100000L)
  while (lRet < (maxNoData - (BUFSIZE * this->noChannels)))
	{
    if (this->verbose)
			  ::printf("WAVE::record() next snd_pcm_readi() %u\n", count); 
    nread = ::snd_pcm_readi(capture_handle, 
                            buf, 
                            BUFSIZE);
    if (this->verbose)
      ::printf("WAVE::record() nread=%d\n", nread);
		if (nread < 0)
		{
      ::fprintf (stderr, "read from audio interface failed (%s)\n",
												::snd_strerror (err));
      /* recover */
      ::snd_pcm_prepare(capture_handle);
    } 
		else 
		{
      // OK here
		  lRet = lRet + (nread * this->noChannels);
			if (this->verbose)
			  ::printf("WAVE::record() no. samples=%ld\n", lRet); 
			this->addSamples(buf, nread * this->noChannels);
    }
    count++;
    //
    // check if the user has hit a key in order to interrupz
    // the recording
    int rc = ::pthread_mutex_lock(&::mutex1);
    if (rc != 0)
    {
       // cannot lock the mutex
       continue;
     }
     localEventMask = ::eventMask;
     ::pthread_mutex_unlock(&::mutex1);
     if (localEventMask & 1)
     {
       if (this->verbose)
         printf("WAVE::record() KEYBOARD interrupt signalled\n");
       break;
     } 
  }   // end while ...
	//
  ::snd_pcm_close (capture_handle);
	//
	//
	zurueck:
  if (this->verbose)
		::printf("WAVE::record() finished lRet=%ld\n", lRet);
  return lRet;
}   // end WAVE::record()   ALSA version

// ----------------------------------------
// Linux/Alsa version
// play a wave file - the name is given as parameter
// Returns: < 0 upon error
int WAVE::play(ALSA *pAlsa,
               unsigned long startMs,
               unsigned long durationMs)  // may be 0
{
  int ret       = 0;
  long lRet     = 0L;
  long noFrames = 0L;
  static short *pShortData = NULL;
  unsigned int nchannels;
  static short *pSampleBuf;
  static snd_pcm_uframes_t lPortion;
  static snd_pcm_t *pHandle;
  char *pError = NULL;
  //
  if (this->verbose)
    ::printf("WAVE::play() start\n");
  if (pAlsa == NULL)
  {
    if (this->verbose)
    {
      ::printf("WAVE::play() ERROR: ALSA instance is NULL\n");
      ::fflush(stdout);
    }
    ret = -1;
    goto zurueck;
  }
  if (this->getInpData() == NULL)
  {
    if (this->verbose)
    {
      ::printf("WAVE::play() ERROR: no inp.data (samples)\n");
      ::fflush(stdout);
    }
    ret = -2;
    goto zurueck;
  }
  //
  pShortData = this->convertSamplesToShort();
  if (this->verbose)
  {
    ::printf("WAVE::play() conversion OK\n");
    ::fflush(stdout);
  }
  //
  noFrames   = this->getNoAllocatedSamples(); //this->noSamples;
  //
  pSampleBuf = pShortData;
  lPortion   = 1024; //pAlsa->getPeriodSize();
  pHandle    = pAlsa->getHandle();
  if (this->verbose)
    ::printf("WAVE::play() lPOrtion=%ld pSampleBuf=%p pHandle=%p\n", 
             lPortion, pSampleBuf, pHandle);
  //
  ::snd_pcm_nonblock(pHandle, 0);   // set blocking mode, not really needed
  //
  while (noFrames > 0L)
  {
    if (this->verbose > 1)
      ::printf("WAVE::play() noFrames=%ld\n", noFrames);
    if (noFrames < lPortion)
      lPortion = noFrames;
    lRet = ::snd_pcm_writei(pHandle, 
                            pSampleBuf, 
                            lPortion);  // no of frames 
    if (this->verbose > 1)
      ::printf("WAVE::play() lRet=%ld\n", lRet);
    switch (lRet)
    {
      case -EBADF:
      {
        if (this->verbose)
         ::printf("WAVE::play() -EBADF\n");
        ret =-10;
        goto zurueck;
      }
      case -EBADFD:
      {
        if (this->verbose)
         ::printf("WAVE::play() -EBADFD\n");
        ret =-11;
        goto zurueck;
      }
      case -ESTRPIPE:
      {
        if (this->verbose)
         ::printf("WAVE::play() -ESTRPIPE\n");
        ret =-12;
        goto zurueck;
      }
      case -EPIPE:
      {
        if (this->verbose)
         ::printf("WAVE::play() -EPIPE\n");
        ::snd_pcm_recover(pAlsa->getHandle(), -EPIPE, 0);
        ret = -13;
        goto zurueck;
      }
      case -5:
      {
        // 
        pError = (char *)(::snd_strerror(-5));
        if (this->verbose)
         ::printf("WAVE::play() ERROR -5 %s\n", pError);
        ret =-14;
        goto zurueck;
      }
      case -22:
      {
        // invalid argument
        pError = (char *)(::snd_strerror(-22));
        if (this->verbose)
         ::printf("WAVE::play() ERROR -22 %s\n", pError);
        ret =-15;
        goto zurueck;
      }
      case -14:
      {
        // bad address
        pError = (char *)(::snd_strerror(-14));
        if (this->verbose)
         ::printf("WAVE::play() ERROR -22 %s\n", pError);
        ret =-16;
        goto zurueck;
      }
    }   // end switch
    //
    if (lRet != lPortion)
    {
      if (this->verbose)
        ::printf("WAVE::play() less data written than requested\n");
      ret = 2;
      goto zurueck;
    }
    noFrames   -= lPortion * this->getNoChannels();
    pSampleBuf += (lRet * this->getNoChannels());
  }   // end while ...
  ret = 1;
  //
  zurueck:
  if (pShortData != NULL)
    delete pShortData;
  if (this->verbose)
    ::printf("WAVE::play() ret=%d\n", ret);
  return ret;
}   // end WAVE::play()   ALSA version
#endif

#ifdef STANDALONE
// --------------------------------------------------------------
void usage()
{
  ::printf("Copyright Dr.E.Huckert 1999-2021 V1.6\n");
  ::printf("Usage: WaveClass10 [-v] [-v2] [-length nnn ]\n");
  ::printf("       [-D device] -i filename\n");
}   // end usage()

// --------------------------------------------------------------
// This is just a test frame fro the WAVE class
// The WAVE routines called here concentrate on some
//     essential methods - not all methods are tested!
int main(int argc, char *argv[])
{
  int   lRet = 0;
  char  fileName[256] = "";
  char alsaDevice[128] = "";
  extern void test1(const char *pFn, const char *pAd);
  extern void test2(const char *pFn, const char *pAd);
  //
  usage();
  if (argc < 3)
  {
    ::printf("ERROR: minimum 3 Arguments needed\n");
    lRet = -99;
    goto zurueck;
  }
  //
  // Evaluate the command line
  for (int n=1; n < argc; n++)
  {
    if (::strcmp(argv[n], "-v") == 0)
      ::verbose = 1;
    if (::strcmp(argv[n], "-v2") == 0)
      ::verbose = 2;
    if (::strcmp(argv[n], "-i") == 0)
      ::strcpy(fileName, argv[n + 1]);
    if (::strcmp(argv[n], "-D") == 0)
      ::strcpy(alsaDevice, argv[n + 1]);
  }
  if (::strlen(fileName) == 0)
  {
    ::printf("main() ERROR: command line not correct (\"-i\" missing?)\n");
    usage();
    lRet = -1;
    goto zurueck;
  }
  #ifndef WIN32
  if (::strlen(alsaDevice) == 0)
  {
    ::printf("main() WARNING: ALSA device=default assumed\n");
    ::strcpy(alsaDevice, "default");
  }
  #endif
  //
  // read and play the WAVE file
  if (::verbose)
    ::printf("main(): start test1\n");
  test1(fileName, alsaDevice);
  if (::verbose)
    ::printf("main(): end test1\n");
  //
  if (::verbose)
    ::printf("\nmain(): start test2\n");
  test2(fileName, alsaDevice);
  if (::verbose)
    ::printf("main(): end test2\n");
  //
  //test3();
  //
  lRet = 0;  // OK
  //
  zurueck:
  #ifndef WIN32
  resetLinuxTerminal();
  #endif
  if (::verbose)
    ::printf("main(): ret=%d\n", (int)lRet);
  return (int)lRet;
}   // end main()
#endif
