// module waveclass10.h
// Copyright Dr.E.Huckert 03-2021

#ifndef WAVECLASS10_INCLUDED
#define WAVECLASS10_INCLUDED

// Versions:
// 14.11.2021 method getSineData() added
// 15.11.2021 method dumpPureData() added
// 25.11.2021 method setBitsSample() added

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <math.h>
// from STL
#include <new>
//#include <string>

#ifdef WIN32
#include <windows.h>
#include <io.h>
#else
#include <string>
#include <unistd.h>
#define DWORD int32_t  // added EH 25.08.2020
#define WORD  int16_t  // added EH 25.08.2020
#endif
#include <sys/stat.h>  // for S_IWRITE

using namespace std;

// file size is typically for 1 second of samples:
//   44 bytes header
//  +   sample rate * block_align
// ex.: 88244 bytes for a mono file at 44100 samples/second = CD quality
typedef struct
{
  DWORD main_chunk;  // "RIFF"
  DWORD mlength;     // Lenth overall without the first 8 Bytes
  DWORD chunk_type;  // "WAVE"
} MAIN_CHUNK;        // 12 bytes

// contains the essential parameters...
typedef struct       // start byte #12
{
  DWORD sub_chunk;   // muss 'fmt ' sein, 4 bytes
  DWORD slength;     // #16, 4 bytes, normally value 16 for PCM
  WORD  format;      // #18, 4 bytes, for PCM = 1
  WORD  channels;    // #20  1=mono, 2=stereo
  DWORD sample_fq;   // #24  4 bytes: 44100
  DWORD bytes_per_second;  // 4 bytes
  WORD  block_align;  // 2 bytes
  WORD  bits_per_sample;   // 2 bytes
} SUB_CHUNK;        //       24 bytes total

typedef struct       // start at pos. (12 + 24)=36
{
  DWORD data_type;   // 'data' or 'LINK' or ...
  DWORD dlength;     // no. of bytes - not samples!
}  DATA_CHUNK;       // 8 bytes

// global (module wide) variables
extern int      verbose;  
static DWORD    actLength    = 0L;

#ifndef WIN32
#define ALSA_PCM_NEW_HW_PARAMS_API
#include <alsa/asoundlib.h>

class WAVE;  // forward decl.

// -----------------------------------------------------------------------
class ALSA
{
  private:
    int verbose;
  protected:
  void basicInit()
  {
    //this->deviceName = "hw:1,0";
    this->deviceName = "default";
    this->sampleRate = 44100;     // preset forCD format
    this->noChannels = 2;         // preset fo stereo
    this->handle     = NULL;
    this->periodSize = 1024;
    this->verbose    = 0;
  };
  //
  public:
  string deviceName;
  unsigned sampleRate;
  unsigned noChannels;
  snd_pcm_t *handle;
  snd_pcm_uframes_t periodSize;
  // constructor assuming a standard device
  ALSA()
  {
    this->basicInit();
  };
  // constructor for a given device
  ALSA(const char *pDeviceName,
       unsigned sr,
       unsigned nc)
  {
    this->basicInit();
    this->deviceName = pDeviceName;
    this->noChannels = nc;
    this->sampleRate = sr;
  };
  ~ALSA();
  int open(WAVE *pWave, int mode);
  int close();
  snd_pcm_t *getHandle()
  { return this->handle; };
  unsigned getNoChannels()
  { return this->noChannels; };
  void setNoChannels(unsigned nc)
  { this->noChannels = nc; };
  unsigned getSampleRate()
  {  return this->sampleRate; };
  snd_pcm_uframes_t getPeriodSize()
  { return this->periodSize; };
  string getDeviceName()
  { return this->deviceName; };
  void setVerbose(int flag)
  { this->verbose = flag; };
};   // end class ALSA
#endif

// ------------------------------------------------------------
class WAVE
{
  private:
    int           ifd;
    int           verbose;
    char          filename[256];
    unsigned long noSamples;     // for stereo: 2 times more than mono!
    SUB_CHUNK     formatHeader; 
    unsigned long samplesSec;    // 44100 for CD quality
    unsigned      bitsSample;    // normally 8=unsigned char or 16=signed short
    unsigned      blockAlign;    // normally 1,2,4
    unsigned      noChannels;    // 1=mono, 2=stereo
    unsigned      bytesSec;
    unsigned      durSecs;
    unsigned      frequency;
    short         format;
    double        *pInpData;   // the samples, always stored as double!!!!!!!!
    unsigned long noAllocatedSamples;
    double        *pSineTable;
    int           sineTabMax;
    unsigned      resCount;
    double        *pSineData;
    double        *pResults;
    int           maxResults;
    int           normalized;  // 1 if in range 1.0 .. -1.0
    double        normMaxVal;
    double        normMinVal;
    //
    void          basicInit();
    //
  public:
    // for generation of data (artificial waves)
    unsigned  freqStart;  // should go into DFT class 
    unsigned  freqEnd;
    unsigned  freqStep;
    //
    WAVE();
    WAVE(unsigned  sampsSec,
         unsigned  ds,
         unsigned  nc);
    WAVE(unsigned  sampsSec,
         unsigned  ds,
         unsigned  nc,
         unsigned  bs);
    ~WAVE();
    void setVerbose(int flag)
    { this->verbose = flag; }; 
    #ifdef WIN32
    int play(HANDLE terminationEvent, 
             unsigned long startMs, 
             unsigned long durMs);
		long record(unsigned long durMs);
    #else
    int play(ALSA *pAlsa, 
             unsigned long startMs, 
             unsigned long durMs);
		long record(ALSA *pAlsa, unsigned long durMs);
    #endif    
    long readMainChunk();
    long readSubChunk();
    long readDataChunk(DATA_CHUNK *pHeader);
    long           read(const char *fileName);
    long           readData(unsigned long *pos);
    unsigned long  normalizeData();
    unsigned long  deNormalizeData();
    void           setDenormalizeRange(double minVal, double maxVal);
    unsigned long  resampleData(float percent);
    int    dumpPureData(long filePos, unsigned long noSamps);
    void   outputData(const char *msg, unsigned processMaxSamples);
    int    makeSineData(unsigned maxX,
                      unsigned noHertz); 
    double *getSineData()
    { return this->pSineData; };
    void   freeSineData();
    void   allocateSampleData(unsigned long noSamp);
    void   freeSampleData();
    short  *convertSamplesToShort();
    unsigned char *convertSamplesToByte();
    long   makeWaveFormSine(unsigned hertz,
                          unsigned long nFrames);
    long   makeWaveFormSquare(unsigned hertz,
                              unsigned long nFrames);
    double processDFTSamples(unsigned long fromSamp,
                             unsigned long noSamps,
                             unsigned noHertz); 
    double processDFT(unsigned long fromMs,
                      unsigned long noMs,
                      unsigned noHertz); 
    double * processInvDFT(unsigned noHertz,
                            unsigned long maxSamples,
                            unsigned *harmonics,
                            unsigned numHarmonics); 
    unsigned long changeAmplitude(int percent);
    void keepResult(double reslt); 
    void findResults(int onlyOne); 
    //
    int  addSamples(short  *pSamp, unsigned long noSamp);
    int  addSamples(double *pSamp, unsigned long noSamp);
    int  addSamples(unsigned char *pSamp, unsigned long noSamp);
    long writeWaveBuffer(const char *fileName);
    long writeMainChunk(int of);
    long writeFormatChunk(int of, unsigned bitsPerSample);
    long writeDataChunk(int ofd, unsigned bitsPerSample); 
    long write(const char *fn);
    long write(const char *fn, unsigned bitsPerSample);
    long writeASCII(const char *fn, unsigned long noSamp);
    void setFrequency(unsigned hertz)
    { this->frequency = hertz; };
    void setDuration(unsigned dur)
    { this->durSecs = dur; };  
    void setSamplesSec(unsigned long val)
    { this->samplesSec = val; };
    unsigned long getSamplesSec()
    { return this->samplesSec; };
    unsigned getBitsSample()
    { return this->bitsSample; };
    void setBitsSample(int nb)
    { this->bitsSample = nb; };
    unsigned getBytesSec()
    { return this->bytesSec; };
    unsigned long getNoSamples()
    { return this->noSamples; };
    void setNoSamples(unsigned long val)
    { this->noSamples = val; };
    unsigned getNoChannels()
    { return this->noChannels; };
    short    getFormat()
    { return this->format; };
    double *getInpData()
    { return this->pInpData; };
    unsigned long getNoAllocatedSamples()
    { return this->noAllocatedSamples; };
    int  getNormalized()
    { return this->normalized; };
    void setNormalized(int norm)
    { this->normalized = norm; };
    bool isSampleMemoryEmpty()
    { 
      if (this->pInpData == NULL) return true;
      else                         return false;
    };
    void protocolDataPortion(unsigned long noOutps,
                             const char *msg);
};   // end class WAVE
#endif
